/******************************************************************************
**
**  COPYRIGHT (C) 2000, 2001, 2002 Intel Corporation.
**
**  The information in this file is furnished for informational use 
**  only, is subject to change without notice, and should not be construed as 
**  a commitment by Intel Corporation. Intel Corporation assumes no 
**  responsibility or liability for any errors or inaccuracies that may appear 
**  in this document or any software that may be provided in association with 
**  this document. 
**
**  FILENAME:       Jflash.cpp
**
**  PURPOSE:        A utility to program Intel flash devices from a PC parallel port.
**
**  LAST MODIFIED:  $Modtime: 2/20/04 10:39a $
******************************************************************************/
/*****************************************************************************
Note: Within these files there are porting instructions that are useful for 
adding additional platform support to this tool. These guides may be located by
searching for the string "PORTING_GUIDE".
*******************************************************************************/



#include <stdio.h>
#include <time.h>
#include <string.h>

#ifdef __linux__
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
//#include <asm/io.h>
#include <sys/io.h>

typedef unsigned long DWORD;
typedef unsigned short WORD;
typedef int INT32;
typedef int BOOL;
const BOOL FALSE=0;
const BOOL TRUE=1;

#define _inp(a) inb(a)
#define _outp(a,d) outb(d,a)
#define _getche() getchar()


#else
#include <windows.h>
#include <conio.h>
#endif /* __linux__ */

#include "Compile_switches.h"
#include "Jflash.h"
#include "jtag.h"
#include "Global_Variables.h"


/*
*******************************************************************************
Forward declarations
*******************************************************************************
*/

int putp(int,int, int); // writes the JTAG data on the parallel port
void id_command(void);	// issues the JTAG command to read the device ID for all 3 chips
//void bypass_all(void);	// issues the JTAG command to put all 3 device in bypass mode
//void extest(void);		// issues the JTAG command EXTEST to the Processor

DWORD access_rom(int, DWORD, DWORD, int);	// Passes read/write/setup data for the Flash memory
DWORD access_bus(int, DWORD, DWORD, int);	// Read/write access to the Processor pins
void Write_Rom(DWORD address, DWORD data);	// Writes to ROM
DWORD Read_Rom(DWORD address);				// Reads from ROM

int test_port(void);	// Looks for and finds a valid parallel port address
int check_id(char*);	// Compares the device IDs for the 3 JTAG chips to expected values
void error_out(char*);	// Prints error and exits program
//void erase_flash(DWORD, DWORD, DWORD, DWORD, int);
void program_flash(DWORD, DWORD, DWORD);
void verify_flash(DWORD, DWORD);
void test_logic_reset(void);
//void test_lock_flash(DWORD, DWORD, DWORD, DWORD, int);
void set_lock_flash(DWORD, DWORD, DWORD, DWORD, int);
void set_address (DWORD);
int PZ_scan_code(int, int, int);
int controller_scan_code(int, int, int);
void pre_DRSCAN(void);
void post_DRSCAN(void);
void pre_IRSCAN(void);
void post_IRSCAN(void);
void mem_rw_mode(int);
void mem_data_driver(int);
void mem_write_enable(int);
void mem_output_enable(int);
void clear_chip_selects(void);
void set_chip_select(DWORD);
DWORD shift_data(int);
void set_data(DWORD);
void jtag_test(void);
void dump_chain(void);
void init_workbuf(void);
void invert_workbuf(void);
void set_pin_chip_select(DWORD);
void gpio_unlock_flash(void);
void gpio_lock_flash(void);
void ParseAndLoad(void);
void ParseAndLoadFlashData(void);
DWORD convert_to_dword(char*);
void AnalyzeChain (void);
void InitPinArray (void);
void InitLockUnlock(void);
void InitAddressOrder(void);
void InitInputDataOrder(void);
void InitOutputDataOrder(void);
void InitChipSelectRegions(void);
void InitFlashGlobals(void);
void UnlockAndEraseBlock(DWORD);
DWORD GetChipSelect(DWORD);
void EraseBlocks(DWORD, DWORD);
void check_file_info(DWORD *fsize , DWORD *last_non_zero, DWORD *last_non_ff, DWORD rom_size);
void check_rom_info(DWORD *max_erase_time, DWORD * dsize, DWORD * max_write_buffer );
int other_bypass(int rp, int ct, int length);
void IR_Command(int command);
void usage(void);
void Set_Platform_Global_Variables(void);
void DeclareDefaults(int);
void Integrity_Check(void);

inline void port_init(int port)
{
#ifdef __linux__
    if (ioperm(port, 3, 1) != 0)
    {
        perror("Couldn't enable port!");
        exit(1);
    }
#endif /* __linux__ */
}

inline void port_finish(int port)
{
#ifdef __linux__
    ioperm(port, 3, 0); // ignore any errors, don't matter
#endif /* __linux__ */
}


/*
*******************************************************************************
*
* FUNCTION:         main
*
* DESCRIPTION:      Entry point for utility
*
* INPUT PARAMETERS: uses optional input parameters:
*						argv[1] = Platform to which the binary file will be downloaded
*                       argv[2] = filename to flash (binary files only)
*                       argv[3] = program options, currently only 'P' for Program
*                       argv[4] = Byte Address     
*						argv[5] = INS or PAR (Insight or parallel cable)
						argv[6] = Debug mode on or off
* RETURNS:          void
*
*******************************************************************************
*/

int main( int argc, char *argv[] )
{
    time_t start;
	DWORD fsize = 0;
	DWORD last_non_zero = 0;
	DWORD last_non_ff = 0;
	DWORD base_address;
    DWORD max_erase_time, dsize, max_write_buffer;
	char base_address_string[12];
	DWORD hexaddress; 
	char plat_name[MAX_IN_LENGTH];
	char buf[10];
	int i = 0;
//	int j;
//	DWORD block_size;


	strcpy(base_address_string, "");
	
	if(argc >= 2)
	{
		//Copy the first argument into a character array
		strcpy(plat_name, argv[1]);
#ifndef __linux__
		// Convert the name of the platform to uppercase
		for(unsigned int i = 0; i < strlen(plat_name); i++)
			plat_name[i] = toupper(plat_name[i]);
#endif

		// use the parameter to construct a filename to parse
		sprintf(data_filename, "%s.dat", plat_name);

		// while we're at it, construct the integrity check filename derived from this.
		sprintf(int_data_filename, "%s_integrity.dat", plat_name);
	}
	else
	{
		if(!UsageShown) usage();
		printf("\nEnter platform data file name: ");
		fgets(plat_name, sizeof(plat_name)-1, stdin);
#ifndef __linux__
		// Convert the name of the platform to uppercase
		for(unsigned int i = 0; i < strlen(plat_name); i++)
			plat_name[i] = toupper(plat_name[i]);
#endif
		// use the parameter to construct a filename to parse
		sprintf(data_filename, "%s.dat", plat_name);

		// while we're at it, construct the integrity check filename derived from this.
		sprintf(int_data_filename, "%s_integrity.dat", plat_name);

	}
	
	char DebugMode[4] = "NOD";			// Either NODebug or DEBug

	if(argc >= 7)
	{
		// Copy the sixth argument into a character array
		strcpy(DebugMode, argv[6]);
		// Convert the parameter to uppercase
		for(int i = 0; i < 4; i++)
			DebugMode[i] = toupper(DebugMode[i]);

		// Compare the mode with list of modes
		if(!strcmp("DEB", DebugMode))
		{
			Debug_Mode = true;
		}
		else if(!strcmp("NOD", DebugMode))
		{
			 Debug_Mode = false;
		}
	}

	// Inhibit questions from being asked. This is to assist in automation scripts

	char AskMe[2] = "A";

	if(argc >= 8)
	{
		// Copy the  argument into a character array
		strcpy(AskMe, argv[7]);
		// Convert the parameter to uppercase
		for(int i = 0; i < 2; i++)
			AskMe[i] = toupper(AskMe[i]);

		// Compare the mode with list of modes
		if(!strcmp("A", AskMe))
		{
			AskQuestions = true;
		}
		else if(!strcmp("D", AskMe))
		{
			 AskQuestions = false;
		}
	}


	ParseAndLoad();
	Set_Platform_Global_Variables();
	InitPinArray();
	InitLockUnlock();
	InitAddressOrder();
	InitInputDataOrder();
	InitOutputDataOrder();
	InitChipSelectRegions();
	AnalyzeChain();


	printf("\nJFLASH Version %s\n",  VERSION);
    printf("COPYRIGHT (C) 2000 - 2003 Intel Corporation\n\n");
	printf("PLATFORM SELECTION:\n");
	printf(" Processor= \t\t%s\n", &WORDARRAY[p_processor][0]);
	printf(" Development System= \t%s\n", &WORDARRAY[p_devsys][0]);
	printf(" Data Version= \t\t%s\n\n", &WORDARRAY[p_dataver][0]);

#ifndef __linux__
    //Test operating system, if WinNT or Win2000 then get device driver handle
	OSVERSIONINFO osvi;
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&osvi);
	if(osvi.dwPlatformId == VER_PLATFORM_WIN32_NT)
	{
        HANDLE h;

		h = CreateFile("\\\\.\\giveio", GENERIC_READ, 0, NULL,
					OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if(h == INVALID_HANDLE_VALUE)
			error_out("Couldn't access giveio device");
		CloseHandle(h);
	}
#endif

	lpt_address = test_port();	// find a valid parallel port address
	if(!lpt_address)
		error_out("Error, unable to find parallel port");

    port_init(lpt_address);

 	// set the extended control register and others for the parallel port
	//lpt_ECR = lpt_address + 0x402;
	//lpt_CTL = lpt_address + 0x02;
	//lpt_STAT = lpt_address + 0x01;
	//#define STATUS_READ _inp(lpt_STAT)
	//#define CONTROL_READ _inp(lpt_CTL)

    test_logic_reset();
 
    // Jtag test appears to have a bug that is preventing use of the Blackstone cable.   
//		{
//			//	jtag_test();  I want everyone to use blackstone so this needs to be out
//		}
	if(argc >= 3)
		strcpy(filename,argv[2]);
	else
		{
			if(!UsageShown) usage();
			printf("Enter binary file name: ");
			fgets(filename, sizeof(filename)-1, stdin);
		}

	char Cable_Type[4] = "PAR";			// Either INS or PAR
	if(argc >= 6)
	{
		// Copy the fifth argument into a character array
		strncpy(Cable_Type, argv[5], 4);
		// Convert the name of the cable to uppercase
		for(int i = 0; i < 4; i++)
			Cable_Type[i] = toupper(Cable_Type[i]);

		// Compare the selected cable name with all supported cables
		if(strncmp("INS", Cable_Type, 4) == 0)
		{
			CableType = Insight_Jtag;
		}
		else if(strncmp("PAR", Cable_Type, 4) == 0)
		{
			 CableType = Parallel_Jtag;
		}
		else if(strncmp("WIG", Cable_Type, 4) == 0)
		{
			 CableType = Wiggler_Jtag;
		}
	}
	test_logic_reset();

	id_command();
	//bypass_all();
	IR_Command(IR_Bypass);
    test_logic_reset();
	gpio_unlock_flash();

    //check_rom_info(&max_erase_time, &dsize, &max_write_buffer, &block_size, &nblocks);
    check_rom_info(&max_erase_time, &dsize, &max_write_buffer);

	if( (in_file = fopen(filename, "rb" )) == NULL)
    {
		error_out("error, can not open binary input file");
    }

	check_file_info(&fsize , &last_non_zero, &last_non_ff, dsize);
    
    // Don't waste time programming ff's at the end of the file this is the erase state
    fsize = last_non_ff;  

	if(argc >= 5)
    {
		sscanf(argv[4],"%x", &hexaddress);
		if(hexaddress & 0x3)
		{
			printf("Start address must be 32 bit aligned!\n");
   			printf("Changing start address to: %x\n",hexaddress & 0xFFFFFFFCul);
   			base_address = (hexaddress & 0xFFFFFFFCul);
   		}	
   			// adjust address for flash addressing mode
   			base_address = (hexaddress & 0xFFFFFFFCul)/ADDR_MULT;

	}
    else
    {
    	base_address = 0;
    }		 


	if(100 - (last_non_zero * 100)/last_non_ff > 20)
	{
		if(AskQuestions)
		{
			printf("The last %2ld percent of image file is all zeros\n",100 - (last_non_zero * 100)/last_non_ff);
			printf("Would you like to save time by not programming that area? [y/n]: ");
			if(toupper(getchar()) == 'Y')
			fsize = last_non_zero;
		}
		else
		{
			fsize = last_non_zero;
		}
	}

	printf("\n");
	char option = 'P';
	if(argc >= 4)
	{
    	option = toupper(*argv[3]);
    }
	
    if(option == 'P')
	{
		EraseBlocks(base_address, fsize);
		program_flash(max_write_buffer, base_address, fsize);
		
		Write_Rom(0, F_READ_ARRAY);				// put back into read mode
		access_rom(READ, base_address, 0x0L, IGNORE_PORT); //extra read to get the pipeline going
		time(&start);
		gpio_lock_flash();

		verify_flash(base_address, fsize);

	}
	else if(option == 'I') // just used for identifying the chips without programming anything
	{
		printf("Program successful completion.\n");
		printf("Processor and Flash Memory Identified.\n");
		printf("No programming operation performed.\n");
	}
	else if(option == 'N') // Program only with no verify cycle
	{
		EraseBlocks(base_address, fsize);
		program_flash(max_write_buffer, base_address, fsize);
		
		Write_Rom(0, F_READ_ARRAY);				// put back into read mode
		access_rom(READ, base_address, 0x0L, IGNORE_PORT); //extra read to get the pipeline going
		time(&start);
		gpio_lock_flash();
	}

	else if(option == 'V')
	{

		Write_Rom(0, F_READ_ARRAY);				// put back into read mode
		access_rom(READ, base_address, 0x0L, IGNORE_PORT); //extra read to get the pipeline going
		time(&start);
		gpio_lock_flash();
		verify_flash(base_address, fsize);
	}
	else if(option == 'S')
	{
		Integrity_Check();
	}

	else if(option == 'T') // test the flash by attempting to write to all sectors
	{
		if(AskQuestions)
		{
			printf("About to test the entire flash memory..... \n");
			printf("Is this what you want to do? (Y or N)\n");
			if(toupper(getchar()) == 'Y')
			{
			i = 0;
			do
			{
				test_logic_reset();
				gpio_unlock_flash();
				base_address = BLOCK_ADDRESS[i];
				EraseBlocks(base_address, fsize);
				program_flash(max_write_buffer, base_address, fsize);
//    			Write_Rom(0, F_READ_ARRAY);				// put back into read mode
//		  		access_rom(READ, base_address, 0x0L, IGNORE_PORT); //extra read to get the pipeline going
//				gpio_lock_flash();
//				verify_flash(base_address, fsize);
				i++;

			} while (BLOCK_ADDRESS[i] != 0);
			i = 0;
			do
			{
				test_logic_reset();
				gpio_unlock_flash();
				base_address = BLOCK_ADDRESS[i];
//				EraseBlocks(base_address, fsize);
//				program_flash(max_write_buffer, base_address, fsize);
    			Write_Rom(0, F_READ_ARRAY);				// put back into read mode
		  		access_rom(READ, base_address, 0x0L, IGNORE_PORT); //extra read to get the pipeline going
				gpio_lock_flash();
				verify_flash(base_address, fsize);
				i++;

			} while (BLOCK_ADDRESS[i] != 0);

				printf("Successfully Completed a write to all sectors\n");
			}
			else
			{
				error_out("Cancelling the test....\n");
			}
		}
		else
		{
			i = 0;
			do
			{
				test_logic_reset();
				gpio_unlock_flash();
				base_address = BLOCK_ADDRESS[i];
				EraseBlocks(base_address, fsize);
				program_flash(max_write_buffer, base_address, fsize);
//    			Write_Rom(0, F_READ_ARRAY);				// put back into read mode
//		  		access_rom(READ, base_address, 0x0L, IGNORE_PORT); //extra read to get the pipeline going
//				gpio_lock_flash();
//				verify_flash(base_address, fsize);
				i++;

			} while (BLOCK_ADDRESS[i] != 0);
			i = 0;
			do
			{
				test_logic_reset();
				gpio_unlock_flash();
				base_address = BLOCK_ADDRESS[i];
//				EraseBlocks(base_address, fsize);
//				program_flash(max_write_buffer, base_address, fsize);
    			Write_Rom(0, F_READ_ARRAY);				// put back into read mode
		  		access_rom(READ, base_address, 0x0L, IGNORE_PORT); //extra read to get the pipeline going
				gpio_lock_flash();
				verify_flash(base_address, fsize);
				i++;

			} while (BLOCK_ADDRESS[i] != 0);


			printf("Successfully Completed a write to all sectors\n");
		}


	}
	else if(option == 'R')
	{

		Write_Rom(0, F_READ_ARRAY);				// put back into read mode
		access_rom(READ, base_address, 0x0L, IGNORE_PORT); //extra read to get the pipeline going
		if((intercom_file_pointer = fopen("VB2JFLASH_READ_DATA.BIN", "w")) == NULL)
		{
			printf("Cannot open output file: %s\n", "VB2JFLASH_READ_DATA.BIN");
			exit(1);
		}
		else
		{ 
			 sprintf(buf, "%X", Read_Rom(base_address)); 
			 fputs(buf,intercom_file_pointer);
			 fclose (intercom_file_pointer); 
			
			//printf("%X\n",Read_Rom(base_address));
		}
	}
	
	else if(option == 'E')
	{
		if(AskQuestions)
		{
			printf("About to erase the entire flash memory..... \n");
			printf("Is this what you want to do? (Y or N)\n");
			if(toupper(getchar()) == 'Y')
			{
				fsize = dsize - 1;
				EraseBlocks(0, fsize);
			}
			else
			{
				error_out("Cancelling the erase....\n");
			}
		}
		else
		{
			fsize = dsize - 1;
			EraseBlocks(0, fsize);
		}
	}
	else
	{
		printf("error specifying programming option. \n\n");
		printf("Commands: \n");
		printf("P = Program and verify\n");
		printf("V = Verify only\n");
		printf("T = Test flash. Program and verify same binary at every block.\n");
		printf("I = Identify processor and flash type. No programming performed.\n");
		printf("R = Read a location and write to file. RESERVED OPERATION\n");
		printf("N = Program only with no verify\n");
		printf("E = Erase entire flash memory\n");
//		printf("S = Stuck bit search. Uses platform_integrity.dat file.\n");
	}

	fclose(in_file);

	test_logic_reset();

    port_finish(lpt_address);
}


/*
*******************************************************************************
*
* FUNCTION:         Set_Platform_Global_Variables
*
* DESCRIPTION:    Evaluates some global variables based on the selected platform             
*                   
*
* INPUT PARAMETERS: none
					
* RETURNS:          None
*
* GLOBAL EFFECTS:   sets the global varibales in Global_Variables.h
*
*******************************************************************************
*/
void Set_Platform_Global_Variables()
{
	
	// Evaluate the following variables based on the selected platform
	ChipSelect0		= convert_to_dword(&WORDARRAY[p_cs0][0]); 	// Global variable for chip select 0
	ChipSelect1		= convert_to_dword(&WORDARRAY[p_cs1][0]); 	// Global variable for chip select 1
	ChipSelect2		= convert_to_dword(&WORDARRAY[p_cs2][0]);	// Global variable for chip select 2
	ChipSelect3		= convert_to_dword(&WORDARRAY[p_cs3][0]);	// Global variable for chip select 3
	ChipSelect4		= convert_to_dword(&WORDARRAY[p_cs4][0]);	// Global variable for chip select 4
	ChipSelect5		= convert_to_dword(&WORDARRAY[p_cs5][0]);	// Global variable for chip select 5
	OutputEnable	= convert_to_dword(&WORDARRAY[p_nOE_OUT][0]);	// Global variable for output enable
	WriteEnable		= convert_to_dword(&WORDARRAY[p_nWE_OUT][0]);	// Global variable for write enable
	MdUpperControl	= convert_to_dword(&WORDARRAY[p_mdupper_ctrl][0]);	// Global variable for MD upper control
	MdLowerControl	= convert_to_dword(&WORDARRAY[p_mdlower_ctrl][0]);	// Global variable for MD lower control
	ReadWriteMode	= convert_to_dword(&WORDARRAY[p_RD_nWR_OUT][0]);	// Global variable for Read Write access mode
	IR_Idcode		= convert_to_dword(&WORDARRAY[p_idcode][0]);	// Global variable for the IDCODE instruction of the IR
	IR_Bypass		= convert_to_dword(&WORDARRAY[p_bypass][0]);	// Global variable for the BYPASS instruction of the IR
	IR_Extest		= convert_to_dword(&WORDARRAY[p_extest][0]);  	// Global variable for the EXTEST instruction of the IR
	ChainLength		= convert_to_dword(&WORDARRAY[p_blength][0]);  	// Global variable for the chain length of the selected platform
	IrLength		= convert_to_dword(&WORDARRAY[p_irlength][0]);  	// Global variable for setting the correct IR length

	if(!strcmp("16", &WORDARRAY[p_datawidth][0] ))
		PlatformIs16bit = true;
	else
		PlatformIs16bit = false;

	// set 16 and 32 bit command and query values

	if (PlatformIs16bit)
	{
	     F_READ_ARRAY = 		0x00FFL;
	     F_READ_IDCODES =   	0x0090L;
	     F_READ_QUERY =     	0x0098L;
	     F_READ_STATUS =    	0x0070L;
	     F_CLEAR_STATUS =   	0x0050L;
	     F_WRITE_BUFFER =   	0x00E8L;
	     F_WORDBYTE_PROG =  	0x0040L;

	     F_BLOCK_ERASE =    	0x0020L;
	     F_BLOCK_ERASE_2ND =	0x00D0L; 

	     F_BLK_ERASE_PS =       0x00B0L;
	     F_BLK_ERASE_PR =       0x00D0L;
	     F_CONFIGURATION =      0x00B8L;

	     F_SET_READ_CFG_REG =   0x0060L;
	     F_SET_READ_CFG_REG_2ND = 0x0003L;

	     F_SET_BLOCK_LOCK =      0x0060L;
	     F_SET_BLOCK_LOCK_2ND =  0x0001L;

	     F_CLEAR_BLOCK_LOCK =    0x0060L;
	     F_CLEAR_BLOCK_LOCK_2ND =0x00D0L;

	     F_PROTECTION =          0x00C0L;

	     F_ATTR_Q =              0x0051L;
	     F_ATTR_R =              0x0052L;
	     F_ATTR_Y =              0x0059L;

	     F_BLOCK_LOCKED =        0x0001L;
	     F_STATUS_READY =        0x0080L;
	}
	else
	{ 
	     F_READ_ARRAY = 		0x00FF00FFL;
	     F_READ_IDCODES =   	0x00900090L;
	     F_READ_QUERY =     	0x00980098L;
	     F_READ_STATUS =    	0x00700070L;
	     F_CLEAR_STATUS =   	0x00500050L;
	     F_WRITE_BUFFER =   	0x00E800E8L;
	     F_WORDBYTE_PROG =  	0x00400040L;

	     F_BLOCK_ERASE =    	0x00200020L;
	     F_BLOCK_ERASE_2ND =	0x00D000D0L; 

	     F_BLK_ERASE_PS =       0x00B000B0L;
	     F_BLK_ERASE_PR =       0x00D000D0L;
	     F_CONFIGURATION =      0x00B800B8L;

	     F_SET_READ_CFG_REG =   0x00600060L;
	     F_SET_READ_CFG_REG_2ND = 0x00030003L;

	     F_SET_BLOCK_LOCK =      0x00600060L;
	     F_SET_BLOCK_LOCK_2ND =  0x00010001L;

	     F_CLEAR_BLOCK_LOCK =    0x00600060L;
	     F_CLEAR_BLOCK_LOCK_2ND =0x00D000D0L;

	     F_PROTECTION =          0x00C000C0L;

	     F_ATTR_Q =              0x00510051L;
	     F_ATTR_R =              0x00520052L;
	     F_ATTR_Y =              0x00590059L;

	     F_BLOCK_LOCKED =        0x00010001L;
	     F_STATUS_READY =        0x00800080L;

	}

	if (PlatformIs16bit)
	{
		ADDR_MULT = 2;
	}
	else
	{
		ADDR_MULT = 4;
	}


}

/*
*******************************************************************************
*
* FUNCTION:         get_rom_info
*
* DESCRIPTION:      get the flash information such as max erase time, flash size, 
*                   max write buffer, block number               
*                   
*
* INPUT PARAMETERS: DWORD *max_erase_time	: max erase time, number of seconds
*					DWORD *dsize			: each flash size, number of 16-bit word
*					DWORD *max_write_buffer : each flash max write buffer, number of 16-bit word
					DWORD *block_size		: each block size, number of 16-bit word
*					DWORD *nblocks			: number of erase blocks in each Flash*
					
* RETURNS:          None
*
* GLOBAL EFFECTS:   None
*
*******************************************************************************
*/
void check_rom_info(DWORD *max_erase_time, DWORD * dsize, DWORD * max_write_buffer)

{	
	DWORD FlashId;
	DWORD FlashIdHigh;
	DWORD FlashIdLow;
	char constructed_string[50];
	char FlashIdString[5];
	bool IDFound = false;
	int IDAttempts = 5;

	do // Check for a recognized Flash ID a few times. Some devices are harder to read
	{
		Write_Rom(0x0, F_READ_ARRAY);	
		Write_Rom(0x0, F_READ_IDCODES);	
	   	
	    if(PlatformIs16bit)
	    {	  	
		  	FlashIdLow = (Read_Rom(0x1) & 0xFFFF);
		  	FlashIdLow = (Read_Rom(0x1) & 0xFFFF); // second read to handle misreads
			FlashId = FlashIdLow;
		}
		else
		{
		  	FlashIdLow = (Read_Rom(0x1) & 0xFFFF);
		  	FlashIdLow = (Read_Rom(0x1) & 0xFFFF);
		    FlashIdHigh = (Read_Rom(0x1) & 0xFFFF0000) >> 16;
		    FlashIdHigh = (Read_Rom(0x1) & 0xFFFF0000) >> 16;
		    if (FlashIdHigh != FlashIdLow) 
		    {
				printf("Upper and Lower flash memory ID does not match.\n");
				printf("You may have a damaged flash memory.\n");
				printf("Upper half reads: %X\n", FlashIdHigh);
				printf("Lower half reads: %X\n", FlashIdLow);
			//	error_out("");
		    }
		    FlashId = FlashIdLow;
		    
		    // determine if the flash is a K3 family and apply the stability fix
		    if ((FlashId == 0x8801) ||
		        (FlashId == 0x8802) ||
		        (FlashId == 0x8803)
		       ) 
		       {
		        K3_STABILITY_FIX_ENABLE = true;
		        if(Debug_Mode) printf("K3 Stability Fix Enabled\n"); 
		       }
		}

		// open the data file for the flash device. If there is no file, then
		// either the flash is not supported, not released, or there was an error
		// reading the Device ID.

		// construct the filename
		sprintf(FlashIdString, "%x", FlashId);
		strcpy(constructed_string, "");

		// construct the processor ID
		strcat(constructed_string, "Flash_");
		strcat(constructed_string, FlashIdString);
		strcat(constructed_string, "_");
		strcat(constructed_string, &WORDARRAY[p_fdevsacross][0]); 
		strcat(constructed_string, "_");
		strcat(constructed_string, &WORDARRAY[p_datawidth][0]);
		strcat(constructed_string, ".dat");

		strcpy(flash_data_filename, constructed_string);  // set the filename
		
		// See if the file exists. If not then try again to confirm the flash ID
		if((flash_file_pointer = fopen(flash_data_filename, "rb")) == NULL)
		{
			printf("Failed to read the Flash ID. Retrying %d more times...\n", IDAttempts -1);
			IDAttempts--;
			if (IDAttempts == 0)
			{
				printf("Cannot open input file: %s\n\n", flash_data_filename);
				printf("This program supports flash devices defined by DAT files\n");
				printf("contained in the same directory as the executable program. \n\n");
				printf("If the file cannot be opened, there are four possibilities:\n\n");
				printf(" 1 - The flash device installed is not supported.\n");
				printf(" 2 - The flash device is a licensed product.\n");
				printf(" 3 - The device ID could not be read, resulting in a poorly\n");
				printf("     constructed filename. The first numeric value in the\n");
				printf("     filename is the device ID. Verify this value with the\n");
				printf("     component specification.\n");
				printf(" 4 - The memory bus is not functional. Check all CPLD and FPGA\n");
				printf("     devices. Make sure that you are using the correct \n");
				printf("     platform data file. \n"); 
				error_out("");

			}
			else
			{
				// hit the system with a JTAG reset and try again
				test_logic_reset();
				// try putting the flash into read mode again
				Write_Rom(0, F_READ_ARRAY);	// put back into read mode
				
				// try forcing asynchronous read mode
				Write_Rom(0x8000, F_SET_READ_CFG_REG);
				Write_Rom(0x8000, F_SET_READ_CFG_REG_2ND);

				
//				Write_Rom(0x0, F_READ_ARRAY);	// put back into read mode
//				Write_Rom(0x0, F_READ_IDCODES);	
			   	
//			  	FlashCR = (Read_Rom(0x5) & 0xFFFF);
//			    FlashCR = (Read_Rom(0x5) & 0xFFFF); // buggy K flash needs another read
	
//				printf("configuration register = %X\n",FlashCR);
				
				Write_Rom(0, F_READ_ARRAY);	// put back into read mode

				access_rom(READ, 0x0L, 0x0L, IGNORE_PORT); //extra read to get the pipeline going
			}
		}
		else
		{
			IDFound = true;	
		}
	 } while(!IDFound);	

	DebugProgress |= foundFlashID;
	ParseAndLoadFlashData();
	InitFlashGlobals();
	printf("Found flash type: %s\n", &FLASHWORDARRAY[pf_type][0]); 

	// set the required variables by using the previousely defined globals

	*max_erase_time = BlockEraseTime;
	*dsize = FlashDeviceSize;
	*max_write_buffer = FlashBufferSize;
}

/*
*******************************************************************************
*
* FUNCTION:         check_file_info
*
* DESCRIPTION:      get the file information and check with rom size
*                   
*
* INPUT PARAMETERS: DWORD *fsize			: file size
					DWORD *last_non_zero	: the point where only 0 or FF remain
					DWORD *last_non_ff		: the point where only ff remain
										
										    if LUBBOCK_SABINAL or
										    LUBBOCK_DBPXA262 is defined,
										    the upper 3 parameters 
											are in 16-bit word number
											else they are in 32-bit DWORD number

					DWORD rom_size			: the rom size in 
*					
					
* RETURNS:          None
*
* GLOBAL EFFECTS:   None
*
*******************************************************************************
*/
void check_file_info(DWORD *fsize , DWORD *last_non_zero, DWORD *last_non_ff, DWORD rom_size)
{	
	
	short li_WORD;
	INT32 li_DWORD;

    if(PlatformIs16bit)
	{
		for(;;)
		{
			int n = fread((WORD *)&li_WORD, sizeof(WORD) , 1, in_file);

    		if(feof(in_file))break; // Any bytes not on a 4 byte boundry at end-of-file will be ignored
			{
				(*fsize)++;
			}
			if(li_WORD != 0 && li_WORD != -1) // Find point in file were only 0's and ff's remain
			{						// For 32 bit data bus, -1 is 0xffffffff, for 16 bit data bus -1 is 0xffff
        		*last_non_zero = *fsize;
			}
			if(li_WORD != -1)  // Find point in file were only ff's remain
			{
        		*last_non_ff = *fsize;
			}
		}

		rewind(in_file);

		// if 16-bit data width used, it assume only one 16-bit flash installed
		if ((*fsize) > rom_size)
			error_out("error, file size is bigger than device size");
	}
	else
	{
		for(;;)
		{
			int n = fread((DWORD *)&li_DWORD, sizeof(DWORD) , 1, in_file);

    		if(feof(in_file))break; // Any bytes not on a 4 byte boundry at end-of-file will be ignored
			{
				(*fsize)++;
			}
			if(li_DWORD != 0 && li_DWORD != -1) // Find point in file were only 0's and ff's remain
			{						// For 32 bit data bus, -1 is 0xffffffff, for 16 bit data bus -1 is 0xffff
        		*last_non_zero = *fsize;
			}
			if(li_DWORD != -1)  // Find point in file were only ff's remain
			{
        		*last_non_ff = *fsize;
			}
		}

		rewind(in_file);

		// if 32-bit data width used. It assume 2 16-bit flash installed. each flash size=fsize
		if ((*fsize) * 2 > rom_size * 2)
			error_out("error, file size is bigger than device size");
	}
}

void jtag_unreset()
{
		_outp(lpt_address+2, 0);				// set all to 0
}

/*
*******************************************************************************
*
* FUNCTION:         putp
*
* DESCRIPTION:      Drives TCK, TDI, and TMS signals and reads TDO signal 
*                   via _outp and _inp calls.
*                   
*                   Cable used had 100 ohm resistors between the following pins
*                   (Cable shipped as part of SA-1110 Development Kit)
*                   Output pins (LPT driving)
*                       LPT D0 Pin 2 and TCK J10 Pin 4
*                       LPT D1 Pin 3 and TDI J10 Pin 11
*                       LPT D2 Pin 4 and TMS J10 Pin 9
*
*                   Input pin (SA-1110 board drives)
*                       LPT Busy Pin 11 and TDO J10 Pin 13
KK pin settings
TDI = Data0
TCK = Data1
TMS = Data2

TDO PaperError

*
*
*
* INPUT PARAMETERS: int tdi - test data in
*
* RETURNS:          int - TDO (Test Data Out)
*
* GLOBAL EFFECTS:   None
*
*******************************************************************************
*/
// Future work: this procedure needs to be cleaned up and extended. There is a 
// strong possibility that that the Altera Byte-Blaster cable could also be
// supported. 

int putp(int tdi, int tms, int rp)
{
	int tdo = -1;

	switch(CableType)
	{
	case Parallel_Jtag:
		{
			// TMS is D2, TDI is D1, and TCK is D0, so construct an output by creating a 
			// rising edge on TCK with TMS and TDI data set.
			_outp(lpt_address, tms*4+tdi*2);				// TCK low 
			_outp(lpt_address, tms*4+tdi*2+1);				// TCK high 

			// if we want to read the port, set TCK low because TDO is sampled on the 
			// TCK falling edge.
			if(rp == READ_PORT)
				_outp(lpt_address, tms*4+tdi*2);			// TCK low
			if(rp == READ_PORT)
				tdo = !((int)_inp(lpt_address + 1) >> 7);	// get TDO data
		} break;

	case Wiggler_Jtag:
		{
			// TMS is D1, TDI is D3, and TCK is D2, so construct an output by creating a 
			// rising edge on TCK with TMS and TDI data set.

			#define WIGGLER_nSRST	0x01	// D0, pin 2
			#define WIGGLER_TMS		0x02	// D1, pin 3
			#define WIGGLER_TCK		0x04	// D2, pin 4
			#define WIGGLER_TDI		0x08	// D3, pin 5
			#define WIGGLER_nTRST	0x10	// D4, pin 6

//			#define WIGGLER_TDO		0x80	// BUSY, pin 11
			#define WIGGLER_TDO_POS	7

			int lpt_data = WIGGLER_nTRST;

			if(tms == 1) lpt_data |= WIGGLER_TMS;
			if(tdi == 1) lpt_data |= WIGGLER_TDI;

   			// construct an output by creating a 
			// rising edge on TCK with TMS and TDI data set.
			lpt_data &= ~WIGGLER_TCK;
			_outp(lpt_address, lpt_data);	// TCK low

			lpt_data |= WIGGLER_TCK;
			_outp(lpt_address, lpt_data);	// TCK high

			// if we want to read the port, set TCK low because TDO is sampled on the 
			// TCK falling edge.
			if(rp == READ_PORT)
			{
				lpt_data &= ~WIGGLER_TCK;
				_outp(lpt_address, lpt_data);	// TCK high ??? Low?
				tdo = !((int)_inp(lpt_address + 1) >> WIGGLER_TDO_POS);	// get TDO data
			}
		} break;

	case Insight_Jtag:
		{
			// There's some bit clearing here that isn't needed. It should make this 
			// code easier to understand.

			//defines for the INSIGHT IJC-1 JTAG cable

			/* the output port (lpt_address) */
			#define INSIGHT_DIN		0x01
			#define INSIGHT_CLK		0x02
			#define INSIGHT_TMS_IN  0x04  
			
			/* Output Enable for the standard JTAG outputs 
			(not TDO since this is an output from the 
			chip we want to talk to */
			#define nINSIGHT_CTRL	0x08
			#define nINSIGHT_PROG	0x10  /* This causes the TDO line to be driven. We'll leave it high.*/


			/*the input port (lpt_address + 1)*/
			#define TDO_INPUT			0x10
			#define TDO_INPUT_BITPOS	4

			int lpt_data;

			//form the data we want to write to the parallel port
			lpt_data = nINSIGHT_PROG;   //Output to TDO off
			lpt_data &= ~nINSIGHT_CTRL;	//Enable the outputs

			if(tms == 1) lpt_data |= INSIGHT_TMS_IN;
			if(tdi == 1) lpt_data |= INSIGHT_DIN;

   			// construct an output by creating a 
			// rising edge on TCK with TMS and TDI data set.
			lpt_data &= ~INSIGHT_CLK;
			_outp(lpt_address, lpt_data);	// TCK low

			lpt_data |= INSIGHT_CLK;
			_outp(lpt_address, lpt_data);	// TCK high

			// if we want to read the port, set TCK low because TDO is sampled on the 
			// TCK falling edge.
			if(rp == READ_PORT)
			{
				lpt_data &= ~INSIGHT_CLK;
				_outp(lpt_address, lpt_data);	// TCK high ??? Low?
				tdo = ((int)_inp(lpt_address + 1) & TDO_INPUT) >> TDO_INPUT_BITPOS;	// get TDO data
			}
		} break;
	}
//	#ifdef DEBUG
//    printf("TDI = %d, TMS = %d, TDO = %d\n",tdi,tms,tdo);
//  #endif

    return tdo;
}
/*
*******************************************************************************
*
* FUNCTION:         id_command
*
* DESCRIPTION:      extract and verify the id codes of the devices in the chain
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/

void id_command(void)
{
    char constructed_string[35];
//    int device;
//   	int bitscount = 0;
    
	IR_Command(IR_Idcode);

    pre_DRSCAN();
	
	strcpy(constructed_string, "");

	// construct the processor ID
	strcat(constructed_string, "**** ");
	strcat(constructed_string, &WORDARRAY[p_proc_id][0]);
	strcat(constructed_string, " ");
	strcat(constructed_string, &WORDARRAY[p_proc_mfg][0]);
	strcat(constructed_string, " ");
	strcat(constructed_string, &WORDARRAY[p_proc_std][0]);

	if(check_id(constructed_string))
		error_out("failed to read device ID for this Platform");
	
	DebugProgress |= foundProcID;
    post_DRSCAN();
}


/*
*******************************************************************************
*
* FUNCTION:         IR_Command
*
* DESCRIPTION:      sends an instruction to the controller and puts all other 
*                   devices into bypass mode.
*
* INPUT PARAMETERS: int	command
*
* RETURNS:          void
*
*******************************************************************************
*/

void IR_Command(int command)
{
    int device;

    pre_IRSCAN();

	for(device = DEVICES_IN_CHAIN -1 ; device >= 0; device--)
	{
		if(device == DEVICE_CONTROLLER)
		{
			if(DEVICEISLAST[device])
			{
				controller_scan_code(command, IGNORE_PORT, TERMINATE);
			}
			else // controller is not the last in the chain
			{
				controller_scan_code(command, IGNORE_PORT, CONTINUE);
			}
		}
		else
		{
			if(DEVICEISLAST[device])
			{
				    other_bypass(IGNORE_PORT, TERMINATE, (int)DEVICEIRLENGTH[device]);
			}
			else 
			{
				    other_bypass(IGNORE_PORT, CONTINUE, (int)DEVICEIRLENGTH[device]);
			}
		} 
	}

    post_IRSCAN();

}


/*
*******************************************************************************
*
* FUNCTION:         access_rom
*
* DESCRIPTION:      This is just an access-bus with the address multiplied by 4
*
* INPUT PARAMETERS: int - rw, or access mode
*                   DWORD - address
*                   DWORD - data
*                   int rp - whether to read or ignore the parallel port
*
* RETURNS:          DWORD - returned data
*
*******************************************************************************
*/
DWORD access_rom(int rw, DWORD address, DWORD data, int rp)
{
	DWORD returnvalue;
    DWORD HalfData = data & FIRST_HALF_WORD_MASK;	// Get the rightmost half of the word only

    // Shift Flash address making A2 the LSB
    
	if(Debug_Mode)
	{
	   	if(PlatformIs16bit)
		 printf("ACCESS_ROM: inp addr = %X, inp HalfData = %X\n", address, HalfData);
		else
		 printf("ACCESS_ROM: inp addr = %X, inp data = %X\n", address, data);
    }

	if(PlatformIs16bit)
		returnvalue = (access_bus(rw, address << 1, HalfData, rp) & FIRST_HALF_WORD_MASK);
    else
		returnvalue = access_bus(rw, address << 2, data, rp);

    if(Debug_Mode)
    	printf("ACCESS_ROM Returns %X\n", returnvalue);

    return returnvalue;
}

/*
*******************************************************************************
*
* FUNCTION:         Write_Rom
*
* DESCRIPTION:      This routine manipulates the memory bus to do reads 
*                   and writes to any memory location.
*
* INPUT PARAMETERS: int address - address to which you need to write to rom
*					int data    - data you need to write to rom
*
* RETURNS:          none
*
*******************************************************************************
*/
void Write_Rom(DWORD address, DWORD data)
{
	if(Debug_Mode)
		printf("Writing to ROM ...\n");

	access_rom(SETUP, address, data, IGNORE_PORT);
	access_rom(WRITE, address, data, IGNORE_PORT);
	access_rom(HOLD, address, data, IGNORE_PORT);
}

/*
*******************************************************************************
*
* FUNCTION:         Read_Rom
*
* DESCRIPTION:      This routine uses access_rom to read from rom
*                   
*
* INPUT PARAMETERS: int address - address from which you need to read from rom
*
* RETURNS:          DWORD - data read
*
*******************************************************************************
*/

DWORD Read_Rom(DWORD address)
{
	if(Debug_Mode)
		printf("Reading from ROM ...\n");
		
		return access_rom(READ, address, 0, READ_PORT);
}


/*
*******************************************************************************
*
* FUNCTION:         access_bus
*
* DESCRIPTION:      This routine manipulates the memory bus to do reads 
*                   and writes to any memory location.
*
* INPUT PARAMETERS: int rw - mode of READ, WRITE, SETUP, HOLD, or RS
*                   DWORD - address of access
*                   DWORD - data to write
*                   int rp - read or ignore port data
*
* RETURNS:          DWORD - retturned data
*
*******************************************************************************
*/


DWORD access_bus(int rw, DWORD address, DWORD data, int rp)
{
	DWORD i;
//	int j;
	int device;

	// Preset SA-1110 or Cotulla pins to default values (all others set in Cotullajtag.h)
    
    clear_chip_selects();
	mem_output_enable(ENABLE);
	mem_write_enable(DISABLE);
	mem_rw_mode(WRITE);
    mem_data_driver(HIZ);
    set_address(address);

    //----------------------------------------------
    if(rw == READ)
	{
		if(Debug_Mode)
        	printf("Read Mode\n");

		mem_rw_mode(READ);
        set_pin_chip_select(address);
	}

    //----------------------------------------------
	else if(rw == WRITE)
	{
		if(Debug_Mode)
        	printf("Write Mode\n");

        mem_write_enable(ENABLE);
		mem_output_enable(DISABLE);
        mem_data_driver(DRIVE);
        set_pin_chip_select(address);
        set_data(data);
	}
    
    //----------------------------------------------
	else if(rw == SETUP || rw == HOLD)	// just like a write except WE, WE needs setup time
	{
		if(Debug_Mode)
        	printf("Setup or Hold Mode\n");

		mem_output_enable(DISABLE);
        mem_data_driver(DRIVE);
		set_data(data);
	}
    //----------------------------------------------
	else if(rw == RS)	// setup prior to RD_nWR_OUT
	{
		if(Debug_Mode)
        	printf("RS Mode\n");

        mem_output_enable(DISABLE);
	}
	else if(rw == K3)
		clear_chip_selects();

    // Common finish

    putp(1,0,IP);	//Run-Test/Idle
	putp(1,0,IP);	//Run-Test/Idle
	putp(1,0,IP);	//Run-Test/Idle
	putp(1,0,IP);	//Run-Test/Idle
	putp(1,1,IP);	//select DR scan
	putp(1,0,IP);	//capture DR
	putp(1,0,IP);	//shift IR ---> (Rami: should be DR?)


	int out_dat[MAX_CHAIN_LENGTH];

	for(device = DEVICES_IN_CHAIN -1 ; device >= 0; device--)
	{
		if(device == DEVICE_CONTROLLER)
		{
			for(i = 1; i < ChainLength; i++)	// shift write data in to JTAG port and read data out
				out_dat[i] = putp(pin[i-1],0,rp);
//			for(i = 0; i < ChainLength; i++)	// shift write data in to JTAG port and read data out
//				out_dat[i] = putp(pin[i],0,rp);

		}
		else
		{
			putp(0,0,IP); // extra clicks for devices in the chain
		} 
	}

#ifdef serious_error
    if(Debug_Mode)
    {
		for(i = 1; i < ChainLength; i++)	
		{
    		if(rw == READ)
			{
				if(i%30 == 0) printf("\n");
				printf(" %d", out_dat[i]);
			}
		}
		printf("\n");
	}
#endif

	putp(0,1,IP);	//Exit1-DR
	putp(1,1,IP);	//Update-DR
	putp(1,0,IP);	//Run-Test/Idle
	putp(1,0,IP);	//Run-Test/Idle
	putp(1,0,IP);	//Run-Test/Idle

	DWORD busdat = 0;

	for(i = 0; i < 32; i++)	// convert serial data to single DWORD
		busdat = busdat | ((DWORD)(out_dat[input_dat_order[i]] << i));
    
    //extest();
	IR_Command(IR_Extest);

	if(Debug_Mode)
	{
    printf("just before return\n");
    dump_chain();
	}

    return(busdat);
}

/*
*******************************************************************************
*
* FUNCTION:         test_port
*
* DESCRIPTION:      Searches for a valid parallel port
*
* INPUT PARAMETERS: void
*
* RETURNS:          int - Address of the port or zero if none available
*
*******************************************************************************
*/

int test_port(void)
{
    int tst = 0;
	// search for valid parallel port

        port_init(LPT1);
		_outp(LPT1, 0x55);
		tst = _inp(LPT1);
        port_finish(LPT1);
		if(tst == 0x55)
        {
            if(Debug_Mode)
            	printf("Parallel Com port found at I/O address: %X\n", LPT1);

            return LPT1;
        }

        port_init(LPT2);
		_outp(LPT2, 0x55);
		tst = _inp(LPT2);
        port_finish(LPT2);
		if(tst == 0x55)
        {
            if(Debug_Mode)
            	printf("Parallel Com port found at I/O address: %X\n", LPT2);
           
            return LPT2;
		}
 
        port_init(LPT3);
		_outp(LPT1, 0x55);
		tst = _inp(LPT3);
        port_finish(LPT3);
		if(tst == 0x55)
        {
            if(Debug_Mode)
            	printf("Parallel Com port found at I/O address: %X\n", LPT3);
            
            return LPT3;
        }

	return(0);	// return zero if none found
}
/*
*******************************************************************************
*
* FUNCTION:         check_id
*
* DESCRIPTION:      Compare an ID string returned from the device with the expected string.
*
* INPUT PARAMETERS: char * device_id - a pointer to the string returned from the device
*
* RETURNS:          int - 0 if ID matches expected, -1 if a match fails
*
*******************************************************************************
*/

int check_id(char *device_id)
{
	// compare passed device ID to the one returned from the ID command
	char in_id[40];
	BOOL error_flag = FALSE;

	if(DEVICES_IN_CHAIN > 1)
		putp(1,0,IGNORE_PORT); // this is a bug fudge factor - look into this
	
	
	for(int i = 34; i >= 0; i--)
	{
		// skip over the spaces in the ID string
        if(i == 4 || i == 21 || i == 33)
		{
			in_id[i] = ' ';
			i--;
		}
        
		if(putp(1,0,READ_PORT) == 0)
			in_id[i] = '0';
		else
			in_id[i] = '1';

		if((in_id[i] != *(device_id + i)) && (*(device_id + i) != '*'))
		{
			error_flag = TRUE;
			
		}
	}
	in_id[35] = 0;

	
	if(error_flag)
	{
		printf("error, failed to read device ID\n");
		printf("check cables and power\n");
		printf("ACT: %s\n",in_id);
		printf("EXP: %s\n\n",device_id);
		return -1;
	}

	int revision =	(int)(in_id[0] - '0') * 8 +
					(int)(in_id[1] - '0') * 4 +
					(int)(in_id[2] - '0') * 2 +
					(int)(in_id[3] - '0');

	switch(revision)	   
	{
		case 0: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID0][0]); 
			break;
		case 1: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID1][0]); 
			break;
		case 2: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID2][0]); 
			break;
		case 3: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID3][0]); 
			break;
		case 4: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID4][0]); 
			break;
		case 5: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID5][0]); 
			break;
		case 6: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID6][0]); 
			break;
		case 7: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID7][0]); 
			break;
		case 8: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID8][0]); 
			break;
		case 9: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID9][0]); 
			break;
		case 10: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID10][0]); 
			break;
		case 11: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID11][0]); 
			break;
		case 12: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID12][0]); 
			break;
		case 13: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID13][0]); 
			break;
		case 14: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID14][0]); 
			break;
		case 15: 
			printf("%s revision %s\n", &WORDARRAY[p_processor][0], &WORDARRAY[p_CID15][0]); 
			break;

		default: printf("Unknown revision number. Out of range!");	// should never get here
	}

	return 0;
}
/*
*******************************************************************************
*
* FUNCTION:         error_out
*
* DESCRIPTION:      generic error printout and program exit.
*
* INPUT PARAMETERS: char * error_string to print before exit
*
* RETURNS:          void
*
* GLOBAL EFFECTS:   Exits the program
*******************************************************************************
*/

void error_out(char *error_string)
{
	printf("%s\n",error_string);
	
	if(Debug_Mode)
	{
		// analyze failure
		printf("\n\nF A I L	U R E    A N A L Y S I S\n\n");
		printf("Processor ID:       %s\n", (DebugProgress & foundProcID) ? "Found" : "Not Found");
		printf("Flash ID:           %s\n", (DebugProgress & foundFlashID) ? "Found" : "Not Found");
		printf("Flash Lock Status:  %s\n", (DebugProgress & unlockedFlash) ? "Unlocked" : "Not Unlocked");
		printf("Verification:       %s\n", (DebugProgress & firstWordVerified) ? "Some Data Verified" : "No Data Verified");
		printf("Flash Erase Status: %s\n\n", (DebugProgress & eraseSuccess) ? "Erased" : "Not Erased");
		
		if(foundProcID & ~DebugProgress)	
		{	
			printf("You have a serious failure.\n");
			printf("The ID of the processor was not detected.\n");
			printf("Check all cable attachments.\n");
			printf("Check for power on Pin 1 of your JTAG connector.\n");
			printf("Check seating of processor and cards in sockets or connectors\n");
			printf("If the bits for the ID string appear to be shifted left or right,\n");
			printf("  then you probably have a JTAG routing switch set in the wrong state.\n");
			printf("If the ACTual bits are all 0s or 1s, you may have a\n");
			printf(   "cable or a power problem on this board.\n");  
		}
		else if(foundFlashID & ~DebugProgress)	
		{	
			printf("You cannot read the flash ID.\n");
			printf("You can read the Processor ID.\n");
			printf("This means that you have some power and cabling looks OK.\n");
			printf("Check all voltages on the flash and processor devices.\n");	
			printf("You have power on the processor chip and at least a few other connections.\n");
			printf("You may have a bad memory bus connection.\n");
			printf("You may be missing a flash data file for this memory type.\n");
			printf("You may have a badly connected flash memory device.\n");
			printf("You may have a bad flash memory.\n");
		}
		else if(unlockedFlash & ~DebugProgress)	
		{	
			printf("You cannot unlock the flash.\n");
			printf("You can read the Processor ID.\n");
			printf("You can read the Flash ID.\n");
			printf("This means that you have some power and cabling looks OK.\n");
			printf("You most likely have a functional memory bus.\n");
			printf("The debug log will show the last Status word from the flash \n");
			printf("  just before this message. Check this value in the Flash Datasheet\n");
			printf("  for clues about the state of the flash.\n"); 
			printf("Check all voltages on the flash and processor devices.\n");	
			printf("You may have a badly connected flash memory device.\n");
			printf("You may have a bad flash memory.\n");
		}
		else if(eraseSuccess & ~DebugProgress)	
		{	
			printf("You are unable to erase the flash.\n");
			printf("You can read the Processor ID.\n");
			printf("You can read the Flash ID.\n");
			printf("You can unlock the flash.\n");
			printf("This means that you have some power and cabling looks OK.\n");
			printf("You most likely have a functional memory bus.\n");
			printf("The debug log will show the last Status word from the flash \n");
			printf("  just before this message. Check this value in the Flash Datasheet\n");
			printf("  for clues about the state of the flash.\n"); 
			printf("Check all voltages on the flash and processor devices.\n");	
			printf("You may have a badly connected flash memory device.\n");
			printf("You may have a bad flash memory.\n");

		}
		else if(firstWordVerified & ~DebugProgress)	
		{	
			printf("You have at least one data word verified.\n");
			printf("You can read the Processor ID.\n");
			printf("You can read the Flash ID.\n");
			printf("You can unlock the flash.\n");
			printf("You can erase the flash.\n");
			printf("This means that you have some power and cabling looks OK.\n");
			printf("You most likely have a functional memory bus.\n");
			printf("You probably got some other verify error.\n");
			printf("Check the actual and expected data values for clues \n");
			printf("  indicating open or shorted data lines.\n");
			printf("Check all voltages on the flash and processor devices.\n");	
			printf("You may have a badly connected flash memory device.\n");
			printf("You may have a bad flash memory.\n");
		}
		else 
		{	
			printf("No data was verified to be programmed.\n");
			printf("You can read the Processor ID.\n");
			printf("You can read the Flash ID.\n");
			printf("You can unlock the flash.\n");
			printf("You can erase the flash.\n");
			printf("This means that you have some power and cabling looks OK.\n");
			printf("You most likely have a functional memory bus.\n");
			printf("You probably did not actually program anything into flash memory.\n");
			printf("Check the actual and expected data values for clues \n");
			printf("  indicating open or shorted data lines.\n");
			printf("Check all voltages on the flash and processor devices.\n");	
			printf("You may have a badly connected flash memory device.\n");
			printf("You may have a bad flash memory.\n");

		}

	}
	exit(1);
}

/*
*******************************************************************************
*
* FUNCTION:         program_flash
*
* DESCRIPTION:      program the flash using buffered writes
*
* INPUT PARAMETERS: DWORD max_write_buffer derived from the flash query
*                  DWORD base_address
*                   DWORD fsize (flash size)
*
* RETURNS:          void
*
*******************************************************************************
*/

void program_flash(DWORD max_write_buffer, DWORD base_address, DWORD fsize)
{
	time_t start = 0;
	time_t now = 0;
	WORD li_WORD;
	DWORD li_DWORD;
	DWORD write_word_count;
//	int bus_width;
//	DWORD Status;
	printf("Starting programming\n");

	if(!strcmp("WORD", &WORDARRAY[P_progmode][0] ))
	{
		printf("Using WORD programming mode...\n");

        DWORD lj;
		for(lj = base_address; lj < fsize + base_address; lj = lj +1)
		{
			
			time(&now);
			if(difftime(now,start) > STATUS_UPDATE)	// Update status every 2 seconds
				{
			   	printf("Writing flash at hex address %8lx, %5.2f%% done    \r"
					,lj * ADDR_MULT ,(float)(lj - base_address)/(float)fsize*100.0);
				fflush(stdout);
				time(&start);
				}

			if(!PlatformIs16bit)
			{
				fread((DWORD *)&li_DWORD, sizeof(DWORD) , 1, in_file);
				access_rom(WRITE, lj, F_WORDBYTE_PROG, IGNORE_PORT); 
				access_rom(HOLD, lj, F_WORDBYTE_PROG, IGNORE_PORT);
				access_rom(WRITE, lj, li_DWORD, IGNORE_PORT); 
				access_rom(HOLD, lj, li_DWORD, IGNORE_PORT);
			}
			else
			{
				fread((WORD *)&li_WORD, sizeof(WORD) , 1, in_file);
				access_rom(WRITE, lj, F_WORDBYTE_PROG, IGNORE_PORT); 
				access_rom(HOLD, lj, F_WORDBYTE_PROG, IGNORE_PORT);
				access_rom(WRITE, lj, li_WORD, IGNORE_PORT); 
				access_rom(HOLD, lj, li_WORD, IGNORE_PORT);
			}
		}
			 	Write_Rom(lj, F_READ_ARRAY);
	}

	else if(!strcmp("BUFFER", &WORDARRAY[P_progmode][0] ))
	{
		printf("Using BUFFER programming mode...\n");

		// "Write Buffer" flow.
		// This uses almost half the cycles required by "word programming" flow
		// Status register is not read to save time.  There is also no checking to see
		// if maximum "Write Buffer Program Time" is violated.  However even with the
		// fastest parallel port bus speed this should not be a problem
		// (i.e. 16 words * 300 JTAG chain length * 4 parallel port cycles * 1uS fast
		// parallel port cycle = 19mS, typical write buffer program times are in the 200uS range).

	if(!PlatformIs16bit)
		write_word_count = (max_write_buffer - 1) + ((max_write_buffer - 1) << 16);
	else
		write_word_count = max_write_buffer - 1;


		time(&start);

        DWORD lj;
		for(lj = base_address; lj < fsize + base_address; lj = lj + max_write_buffer)
		{

			access_rom(WRITE, lj, F_WRITE_BUFFER, IGNORE_PORT); // write buffer command
			access_rom(HOLD, lj, F_WRITE_BUFFER, IGNORE_PORT);

			access_rom(WRITE, lj, write_word_count, IGNORE_PORT); // write word count (max write buffer size)
			access_rom(HOLD, lj, write_word_count, IGNORE_PORT);

			time(&now);
			if(difftime(now,start) > STATUS_UPDATE)	// Update status every 2 seconds
				{
			   	printf("Writing flash at hex address %8lx, %5.2f%% done    \r"
					,lj * ADDR_MULT ,(float)(lj - base_address)/(float)fsize*100.0);
				fflush(stdout);
				time(&start);
				}

			if(!PlatformIs16bit)
			{
				for(DWORD lk = 0; lk < max_write_buffer; lk++)
				{
					fread((DWORD *)&li_DWORD, sizeof(DWORD) , 1, in_file);
					access_rom(WRITE, lj+lk, li_DWORD, IGNORE_PORT);  // Write buffer data
					access_rom(HOLD, lj+lk, li_DWORD, IGNORE_PORT);  // New
				 //	printf("writing %x at address %x\n", li_DWORD, lj+lk);
				}
			}
			else
			{
				for(DWORD lk = 0; lk < max_write_buffer; lk++)
				{
					fread((WORD *)&li_WORD, sizeof(WORD) , 1, in_file);
					access_rom(WRITE, lj+lk, li_WORD, IGNORE_PORT);  // Write buffer data
					access_rom(HOLD, lj+lk, li_WORD, IGNORE_PORT);  // New
				}
			}

				// No need to diferentiate between 16 and 32 bit access_rom functions anymore!
				access_rom(WRITE, 0, F_BLOCK_ERASE_2ND, IGNORE_PORT); // Program Buffer to Flash Confirm
				access_rom(HOLD, 0, F_BLOCK_ERASE_2ND, IGNORE_PORT);  //New
		}
			 	Write_Rom(lj, F_READ_ARRAY);
	}

	printf("\nProgramming done\n");

	rewind(in_file);
}
/*
*******************************************************************************
*
* FUNCTION:         verify_flash
*
* DESCRIPTION:      compares data programmed in flash with the original binary file.
*
* INPUT PARAMETERS: DWORD base_address
*                   DWORD flash_size
*
* RETURNS:          void
*
*******************************************************************************
*/

void verify_flash(DWORD base_address, DWORD fsize)
{
	time_t start, now;
	DWORD li_DWORD, li1_DWORD;
	WORD li_WORD, li1_WORD;

	rewind(in_file);
	printf("Starting Verify\n");
	
	time(&start);

	if(PlatformIs16bit)
	{
		for(DWORD lj = base_address + 1; lj <= fsize + base_address; lj++)
			{
				
				if (lj != base_address) DebugProgress |= firstWordVerified;
				 
				fread((WORD *)&li_WORD, sizeof(WORD) , 1, in_file);
			    // toggle the chip select for K3 flash
			    if (K3_STABILITY_FIX_ENABLE) access_rom(RS, lj, 0x0L, READ_PORT);
				
				li1_WORD = (WORD) Read_Rom(lj);
			
			    if (K3_STABILITY_FIX_ENABLE) Read_Rom(lj);  //hack to fix K3 

				time(&now);
				if(difftime(now,start) > STATUS_UPDATE)	// Update status every 2 seconds
					{
						printf("Verifying flash at hex address %8lx, %5.2f%% done    \r"
						,lj*ADDR_MULT ,(float)(lj - base_address)/(float)fsize*100.0);
						fflush(stdout);
						time(&start);
					}
				if(li_WORD != li1_WORD)
					{
						printf("verify error at address = %lx exp_dat = %lx act_dat = %lx\n",(lj - 1)*ADDR_MULT, li_WORD,li1_WORD);
						error_out("");
					}
			}
	}
	else 
	{
		for(DWORD lj = base_address + 1; lj <= fsize + base_address; lj++)
		{
			if (lj != base_address) DebugProgress |= firstWordVerified;
			
			
			fread((DWORD *)&li_DWORD, sizeof(DWORD) , 1, in_file);
		    // toggle the chip select for K3 flash
		    if (K3_STABILITY_FIX_ENABLE) access_rom(RS, lj, 0x0L, READ_PORT);
			li1_DWORD = Read_Rom(lj);
		    //	li1_DWORD = Read_Rom(lj);
			
			
			time(&now);
			if(difftime(now,start) > STATUS_UPDATE)	// Update status every 2 seconds
				{
				printf("Verifying flash at hex address %8lx, %5.2f%% done    \r"
					,lj*ADDR_MULT ,(float)(lj - base_address)/(float)fsize*100.0);
				fflush(stdout);
				time(&start);
				}
			if(li_DWORD != li1_DWORD)
				{
					printf("verify error at address = %lx exp_dat = %lx act_dat = %lx\n",(lj - 1) *ADDR_MULT,li_DWORD,li1_DWORD);
					error_out("");
				}
		}
	}

	DebugProgress |= lastWordVerified;

	printf("\nVerification successful!                                                    \n");
}
/*
*******************************************************************************
*
* FUNCTION:         test_logic_reset
*
* DESCRIPTION:      initializes the JTAG state machine to a known state
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
 
void test_logic_reset(void)
{
    if(Debug_Mode)
    printf("begin test logic reset\n");
    

    // make sure JTAG HW reset is off
    jtag_unreset();

	// keep TMS set to 1 force a test logic reset
	// no matter where you are in the TAP controller
	for(int i=0; i < 6; ++i)
		putp(1,1,IGNORE_PORT);

    if(Debug_Mode)
    printf("finish test logic reset\n");
    
}

/*
*******************************************************************************
*
* FUNCTION:         set_lock_flash
*
* DESCRIPTION:      sets locks bits in specified block
*
* INPUT PARAMETERS: DWORD base_address of flash
*                   DWORD fsize - size of flash
*                   DWORD block_size - block size of flash
*                   DWORD max_erase_time - used for a timeout 
*                   int block_number - block number of interest
*
* RETURNS:          void
*
*******************************************************************************
*/


void set_lock_flash(DWORD base_address, DWORD fsize, DWORD block_size, DWORD max_erase_time, int block_number)
{
	time_t start, now;

	printf("Starting set block lock bit\n");

	for(DWORD lj = base_address; lj < fsize + base_address; lj = lj + block_size)  // locks only blocks to be programmed
		{

		// Rami: Should this be lj instead of 0 ???
		Write_Rom(0, F_SET_BLOCK_LOCK);			//  block lock bit command
		Write_Rom(lj, F_SET_BLOCK_LOCK_2ND);	// Confirm

		time(&start);
		printf("Erasing block %3d   \r",block_number++);
		while(access_rom(RS, 0, 0, READ_PORT) != F_STATUS_READY)	// Loop until successful status return
			{
			Read_Rom(0);
			time(&now);
			if(difftime(now,start) > max_erase_time + 1)	// Check for status timeout
				error_out("Error, Clear lock timed out");
			}
		}
	printf("Set lock bit done                                           \n");
}

/*
*******************************************************************************
*
* FUNCTION:         set_address
*
* DESCRIPTION:      Loads the address into the address bits
*
* INPUT PARAMETERS: address
*
* RETURNS:          void
*
* GLOBAL EFFECTS:   None
*
* ASSUMPTIONS:      None
*
* CALLS:            None
*
* CALLED BY:        Anyone
*
* PROTOTYPE:        void set_address(unsigned int address);
*
*******************************************************************************
*/
void set_address (DWORD address)
{
    DWORD i;
    
    for (i = 0; i < 26; i++)
    {
        pin[addr_order[i]] = ((address >> i) & 1);
    }
}

/*
*******************************************************************************
*
* FUNCTION:         set_data
*
* DESCRIPTION:      Fills the chain with the data bits
*
* INPUT PARAMETERS: DWORD data
*
* RETURNS:          void
*
*******************************************************************************
*/
void set_data(DWORD data)
{
    DWORD i;
    
    for(i = 0; i < 32; i++)
    {
        pin[dat_order[i]] = ((data >> i) & 1);	// set data pins
    }
}

/*
*******************************************************************************
*
* FUNCTION:         set_pin_chip_select
*
* DESCRIPTION:      Sets chip selects depending on the address and the platform
*
* INPUT PARAMETERS: DWORD address
*
* RETURNS:          void
*
*******************************************************************************
*/

void set_pin_chip_select(DWORD address)
{
	    
	    	 if((address >= CSR_LADDR[0]) && (address < CSR_HADDR[0]))  pin[CSR1] = 0;
		else if((address >= CSR_LADDR[1]) && (address < CSR_HADDR[1]))  pin[CSR2] = 0;
		else if((address >= CSR_LADDR[2]) && (address < CSR_HADDR[2]))  pin[CSR3] = 0;
		else if((address >= CSR_LADDR[3]) && (address < CSR_HADDR[3]))  pin[CSR4] = 0;
		else if((address >= CSR_LADDR[4]) && (address < CSR_HADDR[4]))  pin[CSR5] = 0;
		else if((address >= CSR_LADDR[5]) && (address < CSR_HADDR[5]))  pin[CSR6] = 0;

		pin[p_nsdcas] = 0;
		
}

/*
*******************************************************************************
*
* FUNCTION:         clear_chip_selects
*
* DESCRIPTION:      reset all chip selects
*
* INPUT PARAMETERS: None
*
* RETURNS:          none
*
*******************************************************************************
*/

void clear_chip_selects()
{
    // Preset to default values 
	
		pin[ChipSelect0] = 1;
		pin[ChipSelect1] = 1;
		pin[ChipSelect2] = 1;
		pin[ChipSelect3] = 1;
		pin[ChipSelect4] = 1;
		pin[ChipSelect5] = 1;
		pin[p_nsdcas] = 1;
}
/*
*******************************************************************************
*
* FUNCTION:         mem_output_enable
*
* DESCRIPTION:      enable or disable memory output. This pin is connected to 
*                   the output enables of the memory device.
*
* INPUT PARAMETERS: int - enable or disable
*
* RETURNS:          void
*
*******************************************************************************
*/
void mem_output_enable(int endis)
{
    if (endis == ENABLE)
		pin[OutputEnable] = 0;
    else
		pin[OutputEnable] = 1;
}

/*
*******************************************************************************
*
* FUNCTION:         mem_write_enable
*
* DESCRIPTION:      enable or disable memory writes. This pin is connected to 
*                   the write enables of the memory device.
*
* INPUT PARAMETERS: int - enable or disable
*
* RETURNS:          void
*
*******************************************************************************
*/
void mem_write_enable(int endis)
{
    if (endis == ENABLE)
			pin[WriteEnable] = 0;
    else
			pin[WriteEnable] = 1;
}

/*
*******************************************************************************
*
* FUNCTION:         mem_data_driver
*
* DESCRIPTION:      Sets memory data pins to DRIVE or HIZ. 
*
* INPUT PARAMETERS: int - drive or highz
*
* RETURNS:          void
*
*******************************************************************************
*/

void mem_data_driver(int df)
{
	if (df == DRIVE)
	{
		pin[MdUpperControl] = 1;  
		pin[MdLowerControl] = 1;	// Note for Assabet: MdLowerControl = MdUpperControl 
	}
	else
	{
		pin[MdUpperControl] = 0;
		pin[MdLowerControl] = 0;	// Note for Assabet: MdLowerControl = MdUpperControl
	}
}

/*
*******************************************************************************
*
* FUNCTION:         mem_rw_mode
*
* DESCRIPTION:      Sets memory mode to READ or WRITE. 
*
* INPUT PARAMETERS: int - READ or WRITE
*
* RETURNS:          void
*
*******************************************************************************
*/
void mem_rw_mode(int rw)
{
    if (rw == WRITE)
			pin[ReadWriteMode] = 0;
    else
		    pin[ReadWriteMode] = 1;
}


/*
*******************************************************************************
*
* FUNCTION:         gpio_unlock_flash
*
* DESCRIPTION:      Sets the GPIO bits on the Mainstone that are keeping the 
*                   K3 flash locked. 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void gpio_unlock_flash()
{

	if (UnlockFlashCtrl1 != 9999) 
	pin[UnlockFlashCtrl1] = UnlockFlashCtrl1Lev;

	if (UnlockFlashCtrl2 != 9999) 
	pin[UnlockFlashCtrl2] = UnlockFlashCtrl2Lev;

	if (UnlockFlashCtrl3 != 9999) 
	pin[UnlockFlashCtrl3] = UnlockFlashCtrl3Lev;

	if (UnlockFlashCtrl4 != 9999) 
	pin[UnlockFlashCtrl4] = UnlockFlashCtrl4Lev;

}

/*
*******************************************************************************
*
* FUNCTION:         gpio_lock_flash
*
* DESCRIPTION:      Clears the GPIO bits on the Mainstone to re-lock the 
*                   K3 flash. 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void gpio_lock_flash()
{
	if (UnlockFlashCtrl1 != 9999) 
	pin[UnlockFlashCtrl1] = LockFlashCtrl1Lev;

	if (UnlockFlashCtrl2 != 9999) 
	pin[UnlockFlashCtrl2] = LockFlashCtrl2Lev;

	if (UnlockFlashCtrl3 != 9999) 
	pin[UnlockFlashCtrl3] = LockFlashCtrl3Lev;

	if (UnlockFlashCtrl4 != 9999) 
	pin[UnlockFlashCtrl4] = LockFlashCtrl4Lev;

}

/*
*******************************************************************************
*
* FUNCTION:         pre_IRSCAN
*
* DESCRIPTION:      Sets up the state machine to accept an IR SCAN
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void pre_IRSCAN()
{
    if(Debug_Mode)
    printf("begin pre-IR scan code\n");
    

    putp(1,0,IGNORE_PORT);	//Run-Test/Idle
    putp(1,0,IGNORE_PORT);	//Run-Test/Idle
    putp(1,0,IGNORE_PORT);	//Run-Test/Idle
    putp(1,0,IGNORE_PORT);	//Run-Test/Idle
    putp(1,1,IGNORE_PORT);
    putp(1,1,IGNORE_PORT);	//select IR scan
    putp(1,0,IGNORE_PORT);	//capture IR
    putp(1,0,IGNORE_PORT);	//shift IR
}

/*
*******************************************************************************
*
* FUNCTION:         post_IRSCAN
*
* DESCRIPTION:      Get back to IDLE after scanning in the instruction
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void post_IRSCAN()
{
    if(Debug_Mode)
    printf("begin post-IR scan code\n");
    
   	//putp(1,1,IGNORE_PORT);	//Exit1-IR

	putp(1,1,IGNORE_PORT);	//Update-IR
	putp(1,0,IGNORE_PORT);	//Run-Test/Idle
	putp(1,0,IGNORE_PORT);	//Run-Test/Idle
	putp(1,0,IGNORE_PORT);	//Run-Test/Idle
}

/*
*******************************************************************************
*
* FUNCTION:         pre_DRSCAN
*
* DESCRIPTION:      Sets up the state machine to accept an DR SCAN
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void pre_DRSCAN()
{
    if(Debug_Mode)
    printf("begin pre-DR scan code\n");
    
    
    putp(1,0,IGNORE_PORT);	//Run-Test/Idle
    putp(1,0,IGNORE_PORT);	//Run-Test/Idle
    putp(1,0,IGNORE_PORT);	//Run-Test/Idle
    putp(1,0,IGNORE_PORT);	//Run-Test/Idle
    putp(1,1,IGNORE_PORT);  //select DR scan
    putp(1,0,IGNORE_PORT);	//capture DR
//	putp(1,0,IGNORE_PORT);	//shift DR
}

/*
*******************************************************************************
*
* FUNCTION:         post_DRSCAN
*
* DESCRIPTION:      Get back to IDLE after scanning in the data register
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void post_DRSCAN()
{
    if(Debug_Mode)
    	printf("begin post-DR scan code\n");
    
	
    putp(1,1,IGNORE_PORT);	//Exit1-DR
	putp(1,1,IGNORE_PORT);	//Update-DR
	putp(1,0,IGNORE_PORT);	//Run-Test/Idle
	putp(1,0,IGNORE_PORT);	//Run-Test/Idle
	putp(1,0,IGNORE_PORT);	//Run-Test/Idle
}

/*
*******************************************************************************
*
* FUNCTION:         controller_scan_code
*
* DESCRIPTION:      clocks in a specified cotulla IR scan code
*
* INPUT PARAMETERS: int instruction code
*                   int read or ignore port
*                   int continue or terminate instruction stream
*
* RETURNS:          void
*
*******************************************************************************
*/
// NOTE: I think there is a bug here somewhere.	This procedure seems to work,
// but I think the count is off. Future debugging work required. The problem is
// that a 0x1 should be returned from the instruction register after each 
// valid instruction. I'm not getting this value back, but the instructions 
// appear to be valid because the system works. If this bug can be found, then
// the JTAG_TEST procedure could be used again. 

int controller_scan_code(int code, int rp, int ct)
{
    int i;
    int outval = 0;
    int tms = 0;
 //   int codebit;
    if(Debug_Mode)
    	printf("begin controller scan code = %x\n", code);
    
    
    for (i = 0; i < IrLength; i++)
    {
        if (ct == TERMINATE)
        {
            if (i == IrLength -1)
            {
                tms = 1;
            }
        }
        outval |= putp(((code & (1 << i)) >> i), tms, rp) << (IrLength - i - 1);
	   //	codebit = code & 1<<(IrLength -1 -i);
	   //	outval |= putp(codebit, tms, rp) << (IrLength -1 -i);		

    }
    if(Debug_Mode)
    	printf("Controller IR value: %X\n", outval);
    
    return outval;
}

/*
*******************************************************************************
*
* FUNCTION:         PZ_scan_code
*
* DESCRIPTION:      clocks in a specified PZxxxx IR scan code
*
* INPUT PARAMETERS: int code
*
* RETURNS:          void
*
*******************************************************************************
*/
int PZ_scan_code(int code, int rp, int ct)
{
    int i;
    int outval = 0;
    int tms = 0;
    
    if(Debug_Mode)
    	printf("begin PZ scan code\n");
    
    
    for (i = 0; i < PZ_IRLENGTH; i++)
    {
        if (ct == TERMINATE) 
        {
            if (i == PZ_IRLENGTH -1)
            {
                tms = 1;
            }
        }
        
        outval |= putp(((code & (1 << i)) >> i), tms, rp) << (PZ_IRLENGTH - i - 1);
    }
    return outval;
}

/*
*******************************************************************************
*
* FUNCTION:         other_bypass
*
* DESCRIPTION:      clocks in a bypass scan code for an arbitrary device
*
* INPUT PARAMETERS: int code
*
* RETURNS:          void
*
*******************************************************************************
*/
int other_bypass(int rp, int ct ,int length)
{
    int i;
    int outval = 0;
    int tms = 0;
    int code = 0;

    if(Debug_Mode)
    	printf("Putting Other Device in Bypass\n");
    
    for(i = 0; i < length; i++)
	{
		code |= 1 << i; // construct a universal bypass instruction
	}

    for (i = 0; i < length; i++)
    {
        if (ct == TERMINATE) 
        {
            if (i == length -1)
            {
                tms = 1;
            }
        }
        
        outval |= putp(((code & (1 << i)) >> i), tms, rp) << (length - i - 1);
    }
    return outval;
}

/*
*******************************************************************************
*
* FUNCTION:         jtag_test
*
* DESCRIPTION:      tests the JTAG connection by reading the steady state 
*                   instruction register.
*
* INPUT PARAMETERS: void
*
* RETURNS:          int - 0 if passed
*
*******************************************************************************
*/
void jtag_test()
{
    // set all devices into bypass mode as a safe instruction

    pre_IRSCAN();

		if (controller_scan_code(IR_Bypass, READ_PORT, TERMINATE) != 0x1)
		{
			error_out("Jtag test failure. Check connections and power.\n"); 
		}


    post_IRSCAN();
    
    printf("JTAG Test Passed\n");

}
/*
*******************************************************************************
*
* FUNCTION:         dump_chain
*
* DESCRIPTION:      This is a debug routine that dumps the contents of the 
*                   current boundary chain to the standard I/O.
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void dump_chain()
{
    DWORD addrdat = 0;
    DWORD obusdat = 0;
    DWORD ibusdat = 0;

    int i;
 
    printf("------------------------------------------------------\n");

	for(i = 0; i < 32; i++)	// convert serial data to single DWORD
	{
		obusdat = obusdat | (DWORD)(pin[dat_order[i]] << i);
    }
    printf("Data Bus (Output) = %X\n", obusdat);
    
	for(i = 0; i < 32; i++)	// convert serial data to single DWORD
	{
		ibusdat = ibusdat | (DWORD)(pin[input_dat_order[i]] << i);
	}
    printf("Data Bus (Input) = %X\n", ibusdat);

    for(i = 0; i < 26; i++) // convert address data into single DWORD
    {
        addrdat = addrdat | (DWORD)(pin[addr_order[i]] << i);
    }
    
    
    printf("Address Bus = %X\n", addrdat);
    printf("Flash address is: %X\n",addrdat >> 2);
    
	printf("Chip Select 0 = %d\n", pin[ChipSelect0]);
	
	// Bulverde or Dimebox shortchain cannot read bus data
	if(!PlatformIsBulverdeDimeboxShortChain)
	{
		printf("Chip Select 1 = %d\n", pin[ChipSelect1]);
		printf("Chip Select 2 = %d\n", pin[ChipSelect2]);
		printf("Chip Select 3 = %d\n", pin[ChipSelect3]);
		printf("Chip Select 4 = %d\n", pin[ChipSelect4]);
		printf("Chip Select 5 = %d\n", pin[ChipSelect5]);
	}

	printf("Mem Out Enable      = %d\n", pin[OutputEnable]);    
	printf("Mem Write Enable    = %d\n", pin[WriteEnable]);    
	printf("Mem READ/WRITE Mode = %d\n", pin[ReadWriteMode]);    
	printf("------------------------------------------------------\n");

}   



void usage()
	{
		UsageShown = true;
		printf("Usage:\n\n");
		printf(">JFLASHMM [*PLAT][*IMAGE][P,V,E,T,I,N][ADDR][INS, PAR, WIG][NOD, DEB][A, D]\n\n");
		printf("* = required parameter\n");
		printf("Optional parameters assume the first item in the list.\n\n");
		printf("Where:\n\n");
		printf("[PLAT]	    =The name of the platform data file without the .DAT.\n");
		printf("[IMAGE]	    =The name of the image to program in flash.\n");
		printf("[P,V,E,T,I,N]  where:\n");
		printf("  P = Program and verify. This is the default if no command specified.\n");
		printf("  V = Verify only.\n");
		printf("  E = Erase entire flash device.\n");
		printf("  T = Test every block of flash by programming it. Use 'Testfile.bin.'\n");
		printf("  I = Identify processor step and flash type. No programming done.\n");
		printf("  N = Program the flash but skip the verify. Use for stable systems.\n");
		printf("[ADDR]          =Hex byte address to start. No leading 0X, assumes 0\n");
		printf("[INS, PAR, WIG] =Insight IJC cable, passive parallel or Wiggler\n");
		printf("[NOD, DEB]      =NODebug (Normal) or DEBug mode\n"); 
		printf("[A, D]          =(A)sk or (D)on't ask any questions\n\n");
		printf("Example 1: JFlashmm DBPXA250 MyImage.bin P 0 INS\n");
		printf("Example 2: JFlashmm DBPXA250 MyImage.bin\n");
	}

/*
*******************************************************************************
*
* FUNCTION:         ParseAndLoad
*
* DESCRIPTION:      This procedure parses the data file and loads a global array
*					of strings that wlll be used to construct the rest of the  
*                   variables with the data acquired. The input file is a global.
*					Please refer to the data file for a description of the syntax. 
*
* INPUT PARAMETERS: Void
*
* RETURNS:          void
*
*******************************************************************************
*/
void ParseAndLoad()
{
	char element[132]; /* no word above 132 characters */
	int count = 0;
	unsigned long i = 0;
	bool comment = false;
//	char firstchar[1] = "";

	// initialize the data array strings
	for(i=0; i<=MAX_DATA; i++)
	{
		strcpy(&WORDARRAY[i][0],"");
	}	


	if((data_file_pointer = fopen(data_filename, "rb")) == NULL)
	{
		printf("Cannot open input file: %s\n", data_filename);
		printf("This program supports platforms defined by DAT files\n");
		printf("contained in the directory. Please specify the platform\n");
		printf("by typing the name of the DAT file. Do not type the DAT extension.\n");
		exit(1);
	}
	else
	{ 
		if(Debug_Mode)
			printf("Parsing Platform Data File: %s\n", data_filename);

		// Display the file information stuff
		while( fscanf( data_file_pointer, " %s", element)!=EOF)
		{
			if(!strcmp("/*", element)) // found a begin of comments
			{
				comment = true;
			}
			else if(!strcmp("*/", element)) // found an end of comments
			{
				comment = false;
			}
			else
			{
				if(!comment)
				{
					// load up the data array with the raw strings
				   	strcpy(&WORDARRAY[count][0], element);
					
					if (Debug_Mode)
						printf("WORDARRAY: %d\t%s\n",count,element);
					
					count++;
				}
	    	}// end of else 
		}/* end of while construction */ 
	}// end of else
	
	// Certain data is used for validation of the file
	if(strcmp(VERSION_LOCK, &WORDARRAY[p_verlock][0]))
	{
		printf("Invalid Version Lock string. %s\n", &WORDARRAY[p_verlock][0]);
		printf("Please update your data file for this platform.\n");
		printf("This version of software is not compatible with the data file.\n");
		exit(1);
	}

	else if(strcmp("1111", &WORDARRAY[p_cp1][0]))
	{
		printf("Invalid Checkpoint number 1.\n");
		printf("You have corrupted data for this platform.\n");
		exit(1);
	}

	else if(strcmp("2222", &WORDARRAY[p_cp2][0]))
	{
		printf("Invalid Checkpoint number 2.\n");
		printf("You have corrupted data for this platform.\n");
		exit(1);
	}
	else if(strcmp("3333", &WORDARRAY[p_cp3][0]))
	{
		printf("Invalid Checkpoint number 3.\n");
		printf("You have corrupted data for this platform.\n");
		exit(1);
	}

	fclose(data_file_pointer);
		 
}// ParseAndLoad 

/*
*******************************************************************************
*
* FUNCTION:         ParseAndLoadFlashData
*
* DESCRIPTION:      This procedure parses the flash data file and loads a global array
*					of strings that will be used to construct the rest of the  
*                   variables with the data acquired. The input file is a global.
*					Please refer to the data file for a description of the syntax. 
*
* INPUT PARAMETERS: Void
*
* RETURNS:          void
*
*******************************************************************************
*/
void ParseAndLoadFlashData()
{
	char element[132]; /* no word above 132 characters */
	int count = 0;
	unsigned long i = 0;
	bool comment = false;
//	char firstchar[1] = "";

	// initialize the data array strings
	for(i=0; i<=MAX_FLASH_DATA; i++)
	{
		strcpy(&FLASHWORDARRAY[i][0],"");
	}	


	if((flash_file_pointer = fopen(flash_data_filename, "rb")) == NULL)
	{
		// should never get here because we would not have been called if
		// the file doesn't exist. 
		printf("Flash data file not found\n");
		exit(1);
	}
	else
	{ 
		if(Debug_Mode)
			printf("Parsing Flash Data File: %s\n", flash_data_filename);

		// Display the file information stuff
		while( fscanf( flash_file_pointer, " %s", element)!=EOF)
		{
			if(!strcmp("/*", element)) // found a begin of comments
			{
				comment = true;
			}
			else if(!strcmp("*/", element)) // found an end of comments
			{
				comment = false;
			}
			else
			{
				if(!comment)
				{
					// load up the data array with the raw strings
				   	strcpy(&FLASHWORDARRAY[count][0], element);
					
					if (Debug_Mode)
						printf("FLASHWORDARRAY: %d\t%s\n",count,element);
					
					count++;
				}
	    	}// end of else 
		}/* end of while construction */ 
	}// end of else
	
	// Certain data is used for validation of the file
	if(strcmp(FLASH_VERSION_LOCK, &FLASHWORDARRAY[pf_verlock][0]))
	{
		printf("Invalid Version Lock string. %s\n", &FLASHWORDARRAY[pf_verlock][0]);
		printf("Please update your flash data file for this platform.\n");
		printf("This version of software is not compatible with the data file.\n");
		exit(1);
	}

	else if(strcmp("1111", &FLASHWORDARRAY[pf_cp1][0]))
	{
		printf("Invalid Checkpoint number 1.\n");
		printf("You have corrupted data for this platform.\n");
		exit(1);
	}


	fclose(flash_file_pointer);
		 
}// ParseAndLoad 

/*
*******************************************************************************
*
* FUNCTION:         convert_to_dword
*
* DESCRIPTION:      This procedure is used to convert a string that may be keyed 
*                   to be decimal data, hex data, or octal data into a numeric 
*					value.  
*
* INPUT PARAMETERS: char *
*
* RETURNS:          DWORD
*
*******************************************************************************
*/
DWORD convert_to_dword(char *element)
{
	DWORD return_value = 0;
	DWORD i;
	unsigned long hexword = 0x0;
	unsigned long octword = 0x0;

	// if we are expecting numeric data, then decipher the format

	// if the first character is alpha, then upper case it.
	if (islower(element[0]) || isupper(element[0]))
				element[0] = toupper(element[0]);

			if (element[0] == 'X') // a hex value
			{
				for(i = 0; i< 9; i++) // cannot be larger than 8 hex characters + 1
				{
					//strip off the format character
					element[i] = element[i+1];
				}
				// convert the string to a hex numeric value
				sscanf(element,"%x", &hexword);
				return_value = hexword;
			}
			else if (element[0] == 'O') // an octal value
			{
				for(i = 0; i< 12; i++) // cannot be larger than 11 octal chars + 1
				{
					element[i] = element[i+1];
				}
				// convert the string to an octal value
				sscanf(element,"%o", &octword);
				return_value = octword;
			}
			else 
			{
				// it must be a decimal value
				return_value = atoi(element);
			}

	return return_value;
}

/*
*******************************************************************************
*
* FUNCTION:         AnalyzeChain
* DESCRIPTION:      This procedure constructs a description of the devices on  
*                   the chain. Results are stored in Global variables. 
*					 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void AnalyzeChain ()
{
	bool found_controller = false;
	int i;

	// init the globals
	for (i = 0; i < 5; i++)
	{
		DEVICESTATUS[i] = false;
		DEVICETYPE[i] = false;
		DEVICEIRLENGTH[i] = 0;
		DEVICEISLAST[i] = false;
	}

if (!strcmp("Enabled", &WORDARRAY[p_dev1_stat][0]))	DEVICESTATUS[0] = true;
if (!strcmp("Enabled", &WORDARRAY[p_dev2_stat][0]))	DEVICESTATUS[1] = true;
if (!strcmp("Enabled", &WORDARRAY[p_dev3_stat][0]))	DEVICESTATUS[2] = true;
if (!strcmp("Enabled", &WORDARRAY[p_dev4_stat][0]))	DEVICESTATUS[3] = true;
if (!strcmp("Enabled", &WORDARRAY[p_dev5_stat][0]))	DEVICESTATUS[4] = true;

if (!strcmp("Controller", &WORDARRAY[p_dev1_type][0]))	DEVICETYPE[0] = true;
if (!strcmp("Controller", &WORDARRAY[p_dev2_type][0]))	DEVICETYPE[1] = true;
if (!strcmp("Controller", &WORDARRAY[p_dev3_type][0]))	DEVICETYPE[2] = true;
if (!strcmp("Controller", &WORDARRAY[p_dev4_type][0]))	DEVICETYPE[3] = true;
if (!strcmp("Controller", &WORDARRAY[p_dev5_type][0]))	DEVICETYPE[4] = true;

if (!strcmp("Last", &WORDARRAY[p_dev1_islast][0])) DEVICEISLAST[0] = true;
if (!strcmp("Last", &WORDARRAY[p_dev2_islast][0])) DEVICEISLAST[1] = true;
if (!strcmp("Last", &WORDARRAY[p_dev3_islast][0])) DEVICEISLAST[2] = true;
if (!strcmp("Last", &WORDARRAY[p_dev4_islast][0])) DEVICEISLAST[3] = true;
if (!strcmp("Last", &WORDARRAY[p_dev5_islast][0])) DEVICEISLAST[4] = true;

DEVICEIRLENGTH[0] = convert_to_dword(&WORDARRAY[p_dev1_bits][0]);
DEVICEIRLENGTH[1] = convert_to_dword(&WORDARRAY[p_dev2_bits][0]);
DEVICEIRLENGTH[2] = convert_to_dword(&WORDARRAY[p_dev3_bits][0]);
DEVICEIRLENGTH[3] = convert_to_dword(&WORDARRAY[p_dev4_bits][0]);
DEVICEIRLENGTH[4] = convert_to_dword(&WORDARRAY[p_dev5_bits][0]);

for(i=0; i<5; i++)
{
if(DEVICETYPE[i]) DEVICE_CONTROLLER = i;
}


for(i=0; i<5; i++)
{
	if (DEVICESTATUS[i])
	{
		DEVICES_IN_CHAIN++;
		if(DEVICETYPE[i]) 
		{
			found_controller = true;
		} 

		else if(found_controller)
		{
			DEVICES_AFTER++;	
		}
		else
		{
			DEVICES_BEFORE++;
		}
	}
}

if(Debug_Mode)
{
	printf("Devices_before = %d\n", DEVICES_BEFORE);
	printf("Devices_after  = %d\n", DEVICES_AFTER);
	printf("Device controller at position %d\n", DEVICE_CONTROLLER);

}

}

/*
*******************************************************************************
*
* FUNCTION:         InitPinArray
* DESCRIPTION:      This procedure creates the default pin array by setting  
*                   the initial state of all of the pins. 
*					 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void InitPinArray()
{
	DWORD i;

	// initialize the pin array
	for(i=0; i<ChainLength; i++)
	{
		pin[i] = 0;
	}
	
	// stuff the pin array with the pins we want to be high all of the time

	for(i = p_nh1; i <= p_nh47; i++)
	{
		if(strcmp("9999", &WORDARRAY[i][0]))
		{
			pin[convert_to_dword(&WORDARRAY[i][0])] = 1;
		}		
	}

}
/*
*******************************************************************************
*
* FUNCTION:         InitLockUnlock
* DESCRIPTION:      This procedure sets up any external lock or unlock  
*                   controls for the board.  
*					 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void InitLockUnlock()
{

	// first determine if we have any controls

	if(
		(strcmp("9999", &WORDARRAY[p_unlctl1][0])) ||
		(strcmp("9999", &WORDARRAY[p_unlctl2][0])) ||
		(strcmp("9999", &WORDARRAY[p_unlctl3][0])) ||
		(strcmp("9999", &WORDARRAY[p_unlctl4][0]))
	  )
	  	HASLOCKCONTROLS = true;

	if(HASLOCKCONTROLS)
	{
	    
		UnlockFlashCtrl1 = convert_to_dword(&WORDARRAY[p_unlctl1][0]);		// unlock flash control pin 1
		UnlockFlashCtrl2 = convert_to_dword(&WORDARRAY[p_unlctl2][0]);		// unlock flash control pin 2
		UnlockFlashCtrl3 = convert_to_dword(&WORDARRAY[p_unlctl3][0]);		// unlock flash control pin 3
		UnlockFlashCtrl4 = convert_to_dword(&WORDARRAY[p_unlctl4][0]);		// unlock flash control pin 4

		if(!strcmp("1", &WORDARRAY[p_unlctl1_lev][0]))
		{
		 	UnlockFlashCtrl1Lev = 1;
		 	LockFlashCtrl1Lev = 0;
		} 	
		else
		{
			UnlockFlashCtrl1Lev = 0;
		 	LockFlashCtrl1Lev = 1; 	
		}

		if(!strcmp("1", &WORDARRAY[p_unlctl2_lev][0]))
		{
		 	UnlockFlashCtrl2Lev = 1;
		 	LockFlashCtrl2Lev = 0;
		} 	
		else
		{
			UnlockFlashCtrl2Lev = 0;
		 	LockFlashCtrl2Lev = 1; 	
		}
		if(!strcmp("1", &WORDARRAY[p_unlctl3_lev][0]))
		{
		 	UnlockFlashCtrl3Lev = 1;
		 	LockFlashCtrl3Lev = 0;
		} 	
		else
		{
			UnlockFlashCtrl3Lev = 0;
		 	LockFlashCtrl3Lev = 1; 	
		}
		if(!strcmp("1", &WORDARRAY[p_unlctl4_lev][0]))
		{
		 	UnlockFlashCtrl4Lev = 1;
		 	LockFlashCtrl4Lev = 0;
		} 	
		else
		{
			UnlockFlashCtrl4Lev = 0;
		 	LockFlashCtrl4Lev = 1; 	
		}
	}
}
/*
*******************************************************************************
*
* FUNCTION:         InitAddressOrder
* DESCRIPTION:      This procedure fills the address bit array  
*                    
*					 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/

void InitAddressOrder()
{
	addr_order[0] = convert_to_dword(&WORDARRAY[p_a0][0]);
	addr_order[1] = convert_to_dword(&WORDARRAY[p_a1][0]);
	addr_order[2] = convert_to_dword(&WORDARRAY[p_a2][0]);
	addr_order[3] = convert_to_dword(&WORDARRAY[p_a3][0]);
	addr_order[4] = convert_to_dword(&WORDARRAY[p_a4][0]);
	addr_order[5] = convert_to_dword(&WORDARRAY[p_a5][0]);
	addr_order[6] = convert_to_dword(&WORDARRAY[p_a6][0]);
	addr_order[7] = convert_to_dword(&WORDARRAY[p_a7][0]);
	addr_order[8] = convert_to_dword(&WORDARRAY[p_a8][0]);
	addr_order[9] = convert_to_dword(&WORDARRAY[p_a9][0]);
	addr_order[10] = convert_to_dword(&WORDARRAY[p_a10][0]);
	addr_order[11] = convert_to_dword(&WORDARRAY[p_a11][0]);
	addr_order[12] = convert_to_dword(&WORDARRAY[p_a12][0]);
	addr_order[13] = convert_to_dword(&WORDARRAY[p_a13][0]);
	addr_order[14] = convert_to_dword(&WORDARRAY[p_a14][0]);
	addr_order[15] = convert_to_dword(&WORDARRAY[p_a15][0]);
	addr_order[16] = convert_to_dword(&WORDARRAY[p_a16][0]);
	addr_order[17] = convert_to_dword(&WORDARRAY[p_a17][0]);
	addr_order[18] = convert_to_dword(&WORDARRAY[p_a18][0]);
	addr_order[19] = convert_to_dword(&WORDARRAY[p_a19][0]);
	addr_order[20] = convert_to_dword(&WORDARRAY[p_a20][0]);
	addr_order[21] = convert_to_dword(&WORDARRAY[p_a21][0]);
	addr_order[22] = convert_to_dword(&WORDARRAY[p_a22][0]);
	addr_order[23] = convert_to_dword(&WORDARRAY[p_a23][0]);
	addr_order[24] = convert_to_dword(&WORDARRAY[p_a24][0]);
	addr_order[25] = convert_to_dword(&WORDARRAY[p_a25][0]);


}

/*
*******************************************************************************
*
* FUNCTION:         InitInputDataOrder
* DESCRIPTION:      This procedure fills the input data bit array  
*                    
*					 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void InitInputDataOrder()
{
   input_dat_order[0] = convert_to_dword(&WORDARRAY[p_d0in][0]);
   input_dat_order[1] = convert_to_dword(&WORDARRAY[p_d1in][0]);
   input_dat_order[2] = convert_to_dword(&WORDARRAY[p_d2in][0]);
   input_dat_order[3] = convert_to_dword(&WORDARRAY[p_d3in][0]);
   input_dat_order[4] = convert_to_dword(&WORDARRAY[p_d4in][0]);
   input_dat_order[5] = convert_to_dword(&WORDARRAY[p_d5in][0]);
   input_dat_order[6] = convert_to_dword(&WORDARRAY[p_d6in][0]);
   input_dat_order[7] = convert_to_dword(&WORDARRAY[p_d7in][0]);
   input_dat_order[8] = convert_to_dword(&WORDARRAY[p_d8in][0]);
   input_dat_order[9] = convert_to_dword(&WORDARRAY[p_d9in][0]);
   input_dat_order[10] = convert_to_dword(&WORDARRAY[p_d10in][0]);
   input_dat_order[11] = convert_to_dword(&WORDARRAY[p_d11in][0]);
   input_dat_order[12] = convert_to_dword(&WORDARRAY[p_d12in][0]);
   input_dat_order[13] = convert_to_dword(&WORDARRAY[p_d13in][0]);
   input_dat_order[14] = convert_to_dword(&WORDARRAY[p_d14in][0]);
   input_dat_order[15] = convert_to_dword(&WORDARRAY[p_d15in][0]);
   input_dat_order[16] = convert_to_dword(&WORDARRAY[p_d16in][0]);
   input_dat_order[17] = convert_to_dword(&WORDARRAY[p_d17in][0]);
   input_dat_order[18] = convert_to_dword(&WORDARRAY[p_d18in][0]);
   input_dat_order[19] = convert_to_dword(&WORDARRAY[p_d19in][0]);
   input_dat_order[20] = convert_to_dword(&WORDARRAY[p_d20in][0]);
   input_dat_order[21] = convert_to_dword(&WORDARRAY[p_d21in][0]);
   input_dat_order[22] = convert_to_dword(&WORDARRAY[p_d22in][0]);
   input_dat_order[23] = convert_to_dword(&WORDARRAY[p_d23in][0]);
   input_dat_order[24] = convert_to_dword(&WORDARRAY[p_d24in][0]);
   input_dat_order[25] = convert_to_dword(&WORDARRAY[p_d25in][0]);
   input_dat_order[26] = convert_to_dword(&WORDARRAY[p_d26in][0]);
   input_dat_order[27] = convert_to_dword(&WORDARRAY[p_d27in][0]);
   input_dat_order[28] = convert_to_dword(&WORDARRAY[p_d28in][0]);
   input_dat_order[29] = convert_to_dword(&WORDARRAY[p_d29in][0]);
   input_dat_order[30] = convert_to_dword(&WORDARRAY[p_d30in][0]);
   input_dat_order[31] = convert_to_dword(&WORDARRAY[p_d31in][0]);

}

/*
*******************************************************************************
*
* FUNCTION:         InitOutputDataOrder
* DESCRIPTION:      This procedure fills the output data bit array  
*                    
*					 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void InitOutputDataOrder()
{
   dat_order[0] = convert_to_dword(&WORDARRAY[p_d0out][0]);
   dat_order[1] = convert_to_dword(&WORDARRAY[p_d1out][0]);
   dat_order[2] = convert_to_dword(&WORDARRAY[p_d2out][0]);
   dat_order[3] = convert_to_dword(&WORDARRAY[p_d3out][0]);
   dat_order[4] = convert_to_dword(&WORDARRAY[p_d4out][0]);
   dat_order[5] = convert_to_dword(&WORDARRAY[p_d5out][0]);
   dat_order[6] = convert_to_dword(&WORDARRAY[p_d6out][0]);
   dat_order[7] = convert_to_dword(&WORDARRAY[p_d7out][0]);
   dat_order[8] = convert_to_dword(&WORDARRAY[p_d8out][0]);
   dat_order[9] = convert_to_dword(&WORDARRAY[p_d9out][0]);
   dat_order[10] = convert_to_dword(&WORDARRAY[p_d10out][0]);
   dat_order[11] = convert_to_dword(&WORDARRAY[p_d11out][0]);
   dat_order[12] = convert_to_dword(&WORDARRAY[p_d12out][0]);
   dat_order[13] = convert_to_dword(&WORDARRAY[p_d13out][0]);
   dat_order[14] = convert_to_dword(&WORDARRAY[p_d14out][0]);
   dat_order[15] = convert_to_dword(&WORDARRAY[p_d15out][0]);
   dat_order[16] = convert_to_dword(&WORDARRAY[p_d16out][0]);
   dat_order[17] = convert_to_dword(&WORDARRAY[p_d17out][0]);
   dat_order[18] = convert_to_dword(&WORDARRAY[p_d18out][0]);
   dat_order[19] = convert_to_dword(&WORDARRAY[p_d19out][0]);
   dat_order[20] = convert_to_dword(&WORDARRAY[p_d20out][0]);
   dat_order[21] = convert_to_dword(&WORDARRAY[p_d21out][0]);
   dat_order[22] = convert_to_dword(&WORDARRAY[p_d22out][0]);
   dat_order[23] = convert_to_dword(&WORDARRAY[p_d23out][0]);
   dat_order[24] = convert_to_dword(&WORDARRAY[p_d24out][0]);
   dat_order[25] = convert_to_dword(&WORDARRAY[p_d25out][0]);
   dat_order[26] = convert_to_dword(&WORDARRAY[p_d26out][0]);
   dat_order[27] = convert_to_dword(&WORDARRAY[p_d27out][0]);
   dat_order[28] = convert_to_dword(&WORDARRAY[p_d28out][0]);
   dat_order[29] = convert_to_dword(&WORDARRAY[p_d29out][0]);
   dat_order[30] = convert_to_dword(&WORDARRAY[p_d30out][0]);
   dat_order[31] = convert_to_dword(&WORDARRAY[p_d31out][0]);

}
/*
*******************************************************************************
*
* FUNCTION:         InitChipSelectRegions
* DESCRIPTION:      This procedure sets the global chip select regions  
*                    
*					 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/

void InitChipSelectRegions()
{

	// address ranges
	CSR_LADDR[0] = convert_to_dword(&WORDARRAY[p_m_reg1_low][0]);
	CSR_HADDR[0] = convert_to_dword(&WORDARRAY[p_m_reg1_high][0]);
	CSR_LADDR[1] = convert_to_dword(&WORDARRAY[p_m_reg2_low][0]);
	CSR_HADDR[1] = convert_to_dword(&WORDARRAY[p_m_reg2_high][0]);
	CSR_LADDR[2] = convert_to_dword(&WORDARRAY[p_m_reg3_low][0]);
	CSR_HADDR[2] = convert_to_dword(&WORDARRAY[p_m_reg3_high][0]);
	CSR_LADDR[3] = convert_to_dword(&WORDARRAY[p_m_reg4_low][0]);
	CSR_HADDR[3] = convert_to_dword(&WORDARRAY[p_m_reg4_high][0]);
	CSR_LADDR[4] = convert_to_dword(&WORDARRAY[p_m_reg5_low][0]);
	CSR_HADDR[4] = convert_to_dword(&WORDARRAY[p_m_reg5_high][0]);
	CSR_LADDR[5] = convert_to_dword(&WORDARRAY[p_m_reg6_low][0]);
	CSR_HADDR[5] = convert_to_dword(&WORDARRAY[p_m_reg6_high][0]);

	// the chip select for the range

	CSR1 = GetChipSelect(convert_to_dword(&WORDARRAY[p_m_reg1_cs][0]));
	CSR2 = GetChipSelect(convert_to_dword(&WORDARRAY[p_m_reg2_cs][0]));
	CSR3 = GetChipSelect(convert_to_dword(&WORDARRAY[p_m_reg3_cs][0]));
	CSR4 = GetChipSelect(convert_to_dword(&WORDARRAY[p_m_reg4_cs][0]));
	CSR5 = GetChipSelect(convert_to_dword(&WORDARRAY[p_m_reg5_cs][0]));
	CSR6 = GetChipSelect(convert_to_dword(&WORDARRAY[p_m_reg6_cs][0]));


}
/*
*******************************************************************************
*
* FUNCTION:         GetChipSelect
* DESCRIPTION:      This procedure translates a numeric string to the actual
*					offset for a chip select.  
*                    
*					 
*
* INPUT PARAMETERS: DWORD
*
* RETURNS:          DWORD
*
*******************************************************************************
*/
 

DWORD GetChipSelect(DWORD cs)
{
	
	switch(cs)
		{
			case 0:
				return convert_to_dword(&WORDARRAY[p_cs0][0]);
				break;
			case 1:
				return convert_to_dword(&WORDARRAY[p_cs1][0]);
				break;
			case 2:
				return convert_to_dword(&WORDARRAY[p_cs2][0]);
				break;
			case 3:
				return convert_to_dword(&WORDARRAY[p_cs3][0]);
				break;
			case 4:
				return convert_to_dword(&WORDARRAY[p_cs4][0]);
				break;
			case 5:
				return convert_to_dword(&WORDARRAY[p_cs5][0]);
				break;
			default:
				printf("Illegal chip select chosen!\n");
				printf("Define memory regions with chips selects\n");
				printf("from 0 to 5.\n");
				exit(1);

		}

}

/*
*******************************************************************************
*
* FUNCTION:         InitFlashGlobals
* DESCRIPTION:      This procedure sets the global flash data  
*                    
*					 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void InitFlashGlobals()
{

	REGION_STATUS[0] = false;
	REGION_STATUS[1] = false;
	REGION_STATUS[2] = false;
	REGION_STATUS[3] = false;
	REGION_STATUS[4] = false;
	REGION_STATUS[5] = false;
	REGION_STATUS[6] = false;
	REGION_STATUS[7] = false;
	REGION_STATUS[8] = false;
	REGION_STATUS[9] = false;

	BlockEraseTime = convert_to_dword(&FLASHWORDARRAY[pf_ertime][0]);
	FlashBufferSize = convert_to_dword(&FLASHWORDARRAY[pf_bufsize][0]);
	
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg0status][0]))	REGION_STATUS[0] = true;
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg1status][0]))	REGION_STATUS[1] = true;
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg2status][0]))	REGION_STATUS[2] = true;
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg3status][0]))	REGION_STATUS[3] = true;
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg4status][0]))	REGION_STATUS[4] = true;
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg5status][0]))	REGION_STATUS[5] = true;
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg6status][0]))	REGION_STATUS[6] = true;
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg7status][0]))	REGION_STATUS[7] = true;
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg8status][0]))	REGION_STATUS[8] = true;
	if(!strcmp("enabled", &FLASHWORDARRAY[pf_reg9status][0]))	REGION_STATUS[9] = true;

	REGION_NUM_BLOCKS[0] = convert_to_dword(&FLASHWORDARRAY[pf_reg0number][0]);
	REGION_BLOCKSIZE[0]  = convert_to_dword(&FLASHWORDARRAY[pf_reg0blsize][0]);
	REGION_START_ADDR[0] = convert_to_dword(&FLASHWORDARRAY[pf_reg0start][0]);
	REGION_END_ADDR[0]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg0end][0]);
	REGION_NUM_BLOCKS[1] = convert_to_dword(&FLASHWORDARRAY[pf_reg1number][0]);
	REGION_BLOCKSIZE[1]  = convert_to_dword(&FLASHWORDARRAY[pf_reg1blsize][0]);
	REGION_START_ADDR[1] = convert_to_dword(&FLASHWORDARRAY[pf_reg1start][0]);
	REGION_END_ADDR[1]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg1end][0]);
	REGION_NUM_BLOCKS[2] = convert_to_dword(&FLASHWORDARRAY[pf_reg2number][0]);
	REGION_BLOCKSIZE[2]  = convert_to_dword(&FLASHWORDARRAY[pf_reg2blsize][0]);
	REGION_START_ADDR[2] = convert_to_dword(&FLASHWORDARRAY[pf_reg2start][0]);
	REGION_END_ADDR[2]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg2end][0]);
	REGION_NUM_BLOCKS[3] = convert_to_dword(&FLASHWORDARRAY[pf_reg3number][0]);
	REGION_BLOCKSIZE[3]  = convert_to_dword(&FLASHWORDARRAY[pf_reg3blsize][0]);
	REGION_START_ADDR[3] = convert_to_dword(&FLASHWORDARRAY[pf_reg3start][0]);
	REGION_END_ADDR[3]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg3end][0]);
	REGION_NUM_BLOCKS[4] = convert_to_dword(&FLASHWORDARRAY[pf_reg4number][0]);
	REGION_BLOCKSIZE[4]  = convert_to_dword(&FLASHWORDARRAY[pf_reg4blsize][0]);
	REGION_START_ADDR[4] = convert_to_dword(&FLASHWORDARRAY[pf_reg4start][0]);
	REGION_END_ADDR[4]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg4end][0]);
	REGION_NUM_BLOCKS[5] = convert_to_dword(&FLASHWORDARRAY[pf_reg5number][0]);
	REGION_BLOCKSIZE[5]  = convert_to_dword(&FLASHWORDARRAY[pf_reg5blsize][0]);
	REGION_START_ADDR[5] = convert_to_dword(&FLASHWORDARRAY[pf_reg5start][0]);
	REGION_END_ADDR[5]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg5end][0]);
	REGION_NUM_BLOCKS[6] = convert_to_dword(&FLASHWORDARRAY[pf_reg6number][0]);
	REGION_BLOCKSIZE[6]  = convert_to_dword(&FLASHWORDARRAY[pf_reg6blsize][0]);
	REGION_START_ADDR[6] = convert_to_dword(&FLASHWORDARRAY[pf_reg6start][0]);
	REGION_END_ADDR[6]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg6end][0]);
	REGION_NUM_BLOCKS[7] = convert_to_dword(&FLASHWORDARRAY[pf_reg7number][0]);
	REGION_BLOCKSIZE[7]  = convert_to_dword(&FLASHWORDARRAY[pf_reg7blsize][0]);
	REGION_START_ADDR[7] = convert_to_dword(&FLASHWORDARRAY[pf_reg7start][0]);
	REGION_END_ADDR[7]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg7end][0]);
	REGION_NUM_BLOCKS[8] = convert_to_dword(&FLASHWORDARRAY[pf_reg8number][0]);
	REGION_BLOCKSIZE[8]  = convert_to_dword(&FLASHWORDARRAY[pf_reg8blsize][0]);
	REGION_START_ADDR[8] = convert_to_dword(&FLASHWORDARRAY[pf_reg8start][0]);
	REGION_END_ADDR[8]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg8end][0]);
	REGION_NUM_BLOCKS[9] = convert_to_dword(&FLASHWORDARRAY[pf_reg9number][0]);
	REGION_BLOCKSIZE[9]  = convert_to_dword(&FLASHWORDARRAY[pf_reg9blsize][0]);
	REGION_START_ADDR[9] = convert_to_dword(&FLASHWORDARRAY[pf_reg9start][0]);
	REGION_END_ADDR[9]	 = convert_to_dword(&FLASHWORDARRAY[pf_reg9end][0]);

	// compute FlashDeviceSize. It is the address of the end of the last 
	// block +1.

	if (REGION_STATUS[0]) FlashDeviceSize = REGION_END_ADDR[0] + 1ul; 
	if (REGION_STATUS[1]) FlashDeviceSize = REGION_END_ADDR[1] + 1ul; 
	if (REGION_STATUS[2]) FlashDeviceSize = REGION_END_ADDR[2] + 1ul; 
	if (REGION_STATUS[3]) FlashDeviceSize = REGION_END_ADDR[3] + 1ul; 
	if (REGION_STATUS[4]) FlashDeviceSize = REGION_END_ADDR[4] + 1ul; 
	if (REGION_STATUS[5]) FlashDeviceSize = REGION_END_ADDR[5] + 1ul; 
	if (REGION_STATUS[6]) FlashDeviceSize = REGION_END_ADDR[6] + 1ul; 
	if (REGION_STATUS[7]) FlashDeviceSize = REGION_END_ADDR[7] + 1ul; 
	if (REGION_STATUS[8]) FlashDeviceSize = REGION_END_ADDR[8] + 1ul; 
	if (REGION_STATUS[9]) FlashDeviceSize = REGION_END_ADDR[9] + 1ul; 

	
}

/*
*******************************************************************************
*
* FUNCTION:         EraseBlocks
* DESCRIPTION:      This procedure computes which blocks need to be erased  
*                   by using the file size, the base address, and the memory
*					map of the flash memory. The blocks are then erased. Blocks
*                   are  unlocked  as needed with no warning.  
*					 
*
* INPUT PARAMETERS: base  address, filesize
*
* RETURNS:          void
*
*******************************************************************************
*/
void EraseBlocks(DWORD baseaddress, DWORD filesize)
{
	int region = 0;
	int i;
//	DWORD blockaddress = 0;
//	DWORD regionstart = 0;
//	DWORD regionend = 0;
	DWORD endaddress = 0;
//	bool erasedone = false;
//	bool found_top = false;
	int total_blocks = 0;
	int block_count;
//	bool erasing_on = false;
	int top_block = 0;
	int bottom_block = 0;
    bool askifoffset = false;
	DWORD hexaddress;
    DWORD offsetAddress = 0;
     
    

	endaddress = baseaddress + filesize;
	
	if(Debug_Mode)
		printf("endaddress: %x, baseaddress: %x, filesize: %x FlashDeviceSize: %x\n", endaddress *ADDR_MULT, baseaddress*ADDR_MULT, filesize * ADDR_MULT, FlashDeviceSize);

	if (baseaddress > FlashDeviceSize)
	{
    	printf("Start address is outside of Flash memory\n");
        askifoffset = true;
    }

	if (endaddress > FlashDeviceSize)
    {
   		printf("End of file is outside of Flash memory\n");
        askifoffset = true;
    }
       
    if(askifoffset)
    {
    	printf("Would you like to apply an offset to the flash start address and continue? (Y/N)\n");
		if(toupper(_getche()) == 'Y')
		{
        	printf("Please enter the offset in hex including the 0x.\n");
            scanf("%x", &hexaddress);
            printf ("You entered: %x \n", hexaddress);
            offsetAddress = hexaddress/ADDR_MULT; 
        }
        else
        {
        	error_out("Programming Aborted!");
        }
    }

	baseaddress = baseaddress + offsetAddress;
    endaddress = baseaddress + filesize;
	

	// generate a list of addresses of all of the blocks
	// limit is 512 blocks.

	// init the block address array
	for(i = 0; i < 512; i++)
	{ 
		BLOCK_ADDRESS[i] = 0;
	}

	// get the total number of blocks in this memory space
	
	for(i = 0; i < 10; i++)
	{
		if(REGION_STATUS[i])
		 total_blocks += REGION_NUM_BLOCKS[i];
	}

	// stuff the array with real addresses
	// element 0 must always be 0.
	
	block_count = 0;
	for(region = 0; region < 10; region++)
	{
		for(i=0; i < (int)REGION_NUM_BLOCKS[region]; i++)
		{
			BLOCK_ADDRESS[block_count] = 
				(REGION_START_ADDR[region]/ADDR_MULT) + offsetAddress + (i * REGION_BLOCKSIZE[region]);
			
			block_count++;	
		}		
	}

	if(Debug_Mode)
	{
		for (i= 0; i< 512; i++)
		{
			if(BLOCK_ADDRESS[i] != 0)
			printf("Block number %d is at address %x\n",i, BLOCK_ADDRESS[i]);
		}
	}

	// Erase blocks at the addresses where the input binary will live

	// find the block at the top
	
	for (i = total_blocks -1 ; i >= 0 ; i--)
	{
	   if (BLOCK_ADDRESS[i] <= endaddress)
	   	{
	   		top_block = i;
	   		break;
	   	}	
	}	

	// find the bottom block
	for (i = 0; i <= top_block; i++)
	{
	   if ((BLOCK_ADDRESS[i] <= baseaddress) && (BLOCK_ADDRESS[i+1] > baseaddress ))
	   	{
	   		bottom_block = i;
	   		break;
	   	}
		else // the start address is in the highest block
		{
			bottom_block = top_block;
		}
	}
	
	// erase all of the blocks in between
	for (i = bottom_block; i<= top_block; i++)
	{
		UnlockAndEraseBlock(BLOCK_ADDRESS[i]);
	}	

}
/*
*******************************************************************************
*
* FUNCTION:         UnlockAndEraseBlock
* DESCRIPTION:      This is a support procedure for EraseBlocks. This procedure
*					contains the actual unlock and erase procedure, and 
*					EraseBlocks computes where this procedure is applied.   
*					 
*
* INPUT PARAMETERS: address of block
*
* RETURNS:          void
*
*******************************************************************************
*/
void UnlockAndEraseBlock(DWORD blockaddress)
{

#define allowerase
#ifdef allowerase

	
	time_t start, now;
	DWORD lockstatus;

	// unlock the block
	Write_Rom(blockaddress, F_READ_ARRAY);	
	Write_Rom(blockaddress, F_READ_IDCODES);// Read Identifier Codes


	access_rom(READ, blockaddress + 2, 0, IGNORE_PORT); // read lock configuration (extra read to get JTAG pipeline going)
	lockstatus = Read_Rom(blockaddress + 2);
	if((lockstatus == 0x10001) || (lockstatus == 0x10000) || (lockstatus == 0x00001))
	{
		Write_Rom(blockaddress, F_CLEAR_BLOCK_LOCK); // Clear block lock bit command
		Write_Rom(blockaddress, F_CLEAR_BLOCK_LOCK_2ND);// Confirm

		time(&start);
		printf("Unlocking block at address %x\n", blockaddress * ADDR_MULT);
		while(access_rom(RS, 0, 0, READ_PORT) != F_STATUS_READY)	// Loop until successful status return 0x0080
			{
				Read_Rom(blockaddress);
				time(&now);
				if(difftime(now,start) > BlockEraseTime + 1)	// Check for status timeout
					error_out("\nError, Clear lock timed out");
			}

	}

	DebugProgress |= unlockedFlash;

	// erase the block
	Write_Rom(blockaddress, F_BLOCK_ERASE);		// Erase block command
	Write_Rom(blockaddress, F_BLOCK_ERASE_2ND);	// Erase confirm at the block address

	time(&start);
	printf("Erasing block at address %x   \n",blockaddress * ADDR_MULT);
	while(access_rom(RS, 0, 0, READ_PORT) != F_STATUS_READY)	// Loop until successful status return
		{
			Read_Rom(blockaddress);
			time(&now);
			if(difftime(now,start) > BlockEraseTime + 1)	// Check for erase timeout
				error_out("Error, Block erase timed out");
		}

	DebugProgress |= eraseSuccess;
    Write_Rom(blockaddress, F_READ_ARRAY); // Make sure that the flash is left is read-array mode
#endif // allowerase
}

/*
*******************************************************************************
*
* FUNCTION:         Integrity_Check
* DESCRIPTION:      This procedure walks 1s and 0s through the GPIOs
*					and data lines to look for stuck bits. If there are 
*					other signals that have a bi-directional nature on the
*					JTAG chain, it will be checked here also.    
*					
*					The data file used must be specified on the command line. 
*					 
*
* INPUT PARAMETERS: void
*
* RETURNS:          void
*
*******************************************************************************
*/
void Integrity_Check()
{
	char element[132]; /* no word above 132 characters */
	int count = 0;
	unsigned long i = 0;
	bool comment = false;
//	char firstchar[1] = "";
	unsigned long IBIT; 

	// initialize the data array strings
	for(i=0; i<=1200; i++)
	{
		strcpy(&INT_DATA_ARRAY[i][0],"");
	}	


	if((int_data_file_pointer = fopen(int_data_filename, "rb")) == NULL)
	{
		printf("Cannot open input file: %s\n", data_filename);
		printf("This program supports platforms defined by DAT files\n");
		printf("contained in the directory. Please specify the platform\n");
		printf("by typing the name of the DAT file. Do not type the DAT extension.\n");
		exit(1);
	}
	else
	{ 
		if(Debug_Mode)
			printf("Parsing the Integrity Data File: %s\n", int_data_filename);

		// Display the file information stuff
		while( fscanf( int_data_file_pointer, " %s", element)!=EOF)
		{
			if(!strcmp("/*", element)) // found a begin of comments
			{
				comment = true;
			}
			else if(!strcmp("*/", element)) // found an end of comments
			{
				comment = false;
			}
			else
			{
				if(!comment)
				{
					// load up the data array with the raw strings
				   	strcpy(&INT_DATA_ARRAY[count][0], element);
					
					if (Debug_Mode)
						printf("INT_DATA_ARRAY: %d\t%s\n",count,element);
					
					count++;
				}
	    	}// end of else 
		}/* end of while construction */ 
	}// end of else
	

	fclose(int_data_file_pointer);

	// fill the arrays with data

	count = 0;
	for (i = 0; i<1200; i+=4)
	{
	 
	 strcpy(&INT_NAME_ARRAY[count][0], &INT_DATA_ARRAY[i][0]);
	 INT_OUT_ARRAY[count] 		= convert_to_dword(&INT_DATA_ARRAY[i + 1][0]);
	 INT_IN_ARRAY[count] 		= convert_to_dword(&INT_DATA_ARRAY[i + 2][0]);
	 INT_CONTROL_ARRAY[count] 	= convert_to_dword(&INT_DATA_ARRAY[i + 3][0]);
	 count++;
	}

	if(Debug_Mode)
	{
		printf("Integrity data--------------------\n");
		for (i=0; i<300; i++)
		{
			printf("%s - out=%d, in=%d, ctrl=%d\n", INT_NAME_ARRAY[i], INT_OUT_ARRAY[i], INT_IN_ARRAY[i],INT_CONTROL_ARRAY[i]);
		}
	}
//-------------------------- All data collected, now do the test ------------
/* 
Basic premise of this test: 

Walk through all drivable pins that are outputs, and read the signal on the 
input pin. Ensure that no other pins are set. Repeat the test with setting
all pins high and one low. 

  The procedure should be as follows:

	Define the pin to be set
	Clock the data in
	do an EXTEST instruction to drive the data out
	do a SAMPLE instruction to read the data in
	clock the data out
	check the results

    repeat procedure above for all drivable pins

*/
	test_logic_reset();

	for(IBIT = 35; IBIT < 300; IBIT++)
	{
		if(INT_OUT_ARRAY[IBIT] != 9999)
		printf("Checking %s\n",INT_NAME_ARRAY[IBIT]);
		{
			// define the pin to set
			// first clear all other bits
			for (i=0; i<300; i++)
			{
				if(INT_OUT_ARRAY[i] != 9999)
				{
					pin[INT_OUT_ARRAY[i]] = 0;
				}
			}
			// now set the test bit
			pin[INT_OUT_ARRAY[IBIT]] = 1;
			pin[INT_CONTROL_ARRAY[IBIT]] = 1;

			// clock the data in
		    putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,1,IP);	//select DR scan
			putp(1,0,IP);	//capture DR
			putp(1,0,IP);	//shift DR
			
			for(i = 0; i < ChainLength; i++)
			{
				out_dat[i] = putp(pin[i],0,IP);
			}
			putp(0,1,IP);	//Exit1-DR
			putp(1,1,IP);	//Update-DR
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle

			// EXTEST the data
			IR_Command(IR_Extest);

			// SAMPLE the data
			//IR_Command(1);	// usually the command is "1" but should be defined in the dat
			// Clamp the data
			//IR_Command(4);	// usually the command is "4" but should be defined in the dat



			// clock the data out
		    putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,1,IP);	//select DR scan
			putp(1,0,IP);	//capture DR
			putp(1,0,IP);	//shift DR			
			for(i = 0; i < ChainLength; i++)
			{
				out_dat[i] = putp(pin[i],0,READ_PORT);
			}
			putp(0,1,IP);	//Exit1-DR
			putp(1,1,IP);	//Update-DR
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle

			// EXTEST the data
			//IR_Command(IR_Extest);

			// clock the data out
		    putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,1,IP);	//select DR scan
			putp(1,0,IP);	//capture DR
			putp(1,0,IP);	//shift DR			
			for(i = 0; i < ChainLength; i++)
			{
				out_dat[i] = putp(pin[i],0,READ_PORT);
			}
			putp(0,1,IP);	//Exit1-DR
			putp(1,1,IP);	//Update-DR
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle
			putp(1,0,IP);	//Run-Test/Idle

			count = 0;
			for (i=0; i< 300; i++)
			{
				printf("%s %d %d | ", INT_NAME_ARRAY[i], pin[INT_OUT_ARRAY[i]], out_dat[INT_IN_ARRAY[i]]);
				count ++;
				if (count == 5)
				{
					printf("\n");
					count = 0;
				}
			}
			count = 0;

			for (i=0; i<ChainLength; i++)
			{
				printf("%d-%d-%d | ",i,pin[i],out_dat[i]);
				count ++;
				if (count == 10)
				{
					printf("\n");
					count = 0;
				}

			}

			// check the pin data for integrity
			for (i = 35; i < 300; i++)
			{
				if (INT_IN_ARRAY[i] != 9999)
				{
					if(i != IBIT)
					{
						if(out_dat[INT_IN_ARRAY[i]] != 0)
						{							printf("ERROR: Pin %s stuck high.\n",INT_NAME_ARRAY[i]);
							exit(1);
						}
					}
					else
					{
						if(out_dat[INT_IN_ARRAY[i]] != 1)
						{
							printf("ERROR: Pin %s stuck low.\n",INT_NAME_ARRAY[i]);
							exit(1);
						}
				 	}
				}
			}
		}
	}
}
















